import sqlalchemy as db
from sqlalchemy.exc import IntegrityError
from sqlalchemy.orm import sessionmaker, relationship, backref

import so3g
from spt3g import core as spt3g_core
import datetime as dt
import os
import re
from tqdm import tqdm
import numpy as np
import yaml
import ast
from collections import namedtuple
from enum import Enum

import logging

logger = logging.getLogger(__name__)

from .. import core
from . import load as io_load
from .datapkg_utils import load_configs, walk_files, just_suprsync
from .g3thk_db import G3tHk, HKFiles, HKAgents, HKFields
from .g3thk_utils import pysmurf_monitor_control_list

from sotodlib.io.g3tsmurf_db import (
    Base,
    TimeCodes,
    SupRsyncType,
    Finalize,
    Observations,
    Tags,
    Files,
    Tunes,
    TuneSets,
    ChanAssignments,
    Channels,
    FrameType,
    Frames,
)

Session = sessionmaker()
num_bias_lines = 16


"""
Actions used to define when observations happen
Could be expanded to other Action Based Indexing as well
Strings must be unique, in that they must only show up when they should be used
as observations
"""
SMURF_ACTIONS = {
    "observations": [
        "take_stream_data",
        "stream_data_on",
        "take_noise_psd",
        "take_g3_data",
        "stream_g3_on",
    ],
    "calibrations": ["take_iv", "take_bias_steps", "take_bgmap", "take_noise"],
}


# Types of Frames we care about indexing
type_key = ["Observation", "Wiring", "Scan"]


class TimingParadigm(Enum):
    G3Timestream = 1
    SmurfUnixTime = 2
    TimingSystem = 3
    Mixed = 4


def get_sample_timestamps(frame):
    """
    Gets timestamps of samples in a G3Frame. This will try to get highest
    precision first and move to lower precision methods if that fails.

    Args
    ------
        frame (spt3g_core.G3Frame):
            A G3Frame(Scan) containing streamed detector data.

    Returns
    ---------
        times (np.ndarray):
            numpy array containing timestamps in seconds
        paradigm (TimingParadigm):
            Paradigm used to calculate timestamps.
    """
    logger.warning("get_sample_timestamps is deprecated, how did you get here?")
    if "primary" in frame.keys():
        if False:
            # Do high precision timing calculation here when we have real data
            pass
        else:
            # Try to calculate the timestamp based on the SmurfProcessor's
            # "UnixTime" and the G3Timestream start time.  "UnixTime" is a
            # 32-bit nanosecond clock that steadily increases mod 2**32.
            unix_times = np.array(frame["primary"]["UnixTime"])
            for i in np.where(np.diff(unix_times) < 0)[0]:
                # This corrects for any wrap around
                unix_times[i + 1 :] += 2**32
            times = (
                frame["data"].start.time / spt3g_core.G3Units.s
                + (unix_times - unix_times[0]) / 1e9
            )

            return times, TimingParadigm.SmurfUnixTime
    else:
        # Calculate timestamp based on G3Timestream.times(). Note that this
        # only uses the timestream start and end time, and assumes samples are
        # equispaced.
        times = np.array([t.time / spt3g_core.G3Units.s for t in frame["data"].times()])
        return times, TimingParadigm.G3Timestream


def _file_has_end_frames(filename):
    ended = False

    ## Search through file looking for stream closeout
    reader = so3g.G3IndexedReader(filename)
    logger.debug(f"Reading {filename} to look for observation end.")
    while True:
        frames = reader.Process(None)
        if not frames:
            break
        frame = frames[0]
        if frame.type == spt3g_core.G3FrameType.Observation:
            if frame.get("stream_placement") == "end":
                ended = True
                break
        if frame.type == spt3g_core.G3FrameType.Wiring:
            ## ignore dump frames, they lie (and are at the beginning of observations)
            if frame["dump"]:
                continue
            if "AMCc.SmurfProcessor.FileWriter.IsOpen" in frame["status"]:
                status = {}
                status.update(yaml.safe_load(frame["status"]))
                if not status["AMCc.SmurfProcessor.FileWriter.IsOpen"]:
                    ended = True
                    break
            if "AMCc.SmurfProcessor.SOStream.open_g3stream" in frame["status"]:
                status = {}
                status.update(yaml.safe_load(frame["status"]))
                if not status["AMCc.SmurfProcessor.SOStream.open_g3stream"]:
                    ended = True
                    break
    return ended


class G3tSmurf:
    def __init__(
        self,
        archive_path,
        db_path=None,
        meta_path=None,
        echo=False,
        db_args={},
        finalize={},
        hk_db_path=None,
        make_db=False,
    ):
        """
        Class to manage a smurf data archive.

        Args
        -----
            archive_path: path
                Path to the data directory
            db_path: path, optional
                Path to the sqlite file. Defaults to
                ``<archive_path>/frames.db``
            meta_path: path, optional
                Path of directory containing smurf related metadata (ie. channel
                assignments). Required for full functionality.
            echo: bool, optional
                If true, all sql statements will print to stdout.
            db_args: dict, optional
                Additional arguments to pass to sqlalchemy.create_engine
            finalize: dict, optional
                Arguments required for fast tracking of data file transfers
            hk_db_path: str
                Path the HK database for finalization tracking
            make_db: bool
                if True and db_path does not exist it will make a new database.
                otherwise will throw and error if database path does not exist
        """
        if db_path is None:
            db_path = os.path.join(archive_path, "frames.db")
        self.archive_path = archive_path
        self.meta_path = meta_path
        self.db_path = db_path
        self.hk_db_path = hk_db_path
        self.HK = None
        self.finalize = finalize

        if os.path.exists(self.db_path):
            new_db = False
        elif make_db:
            new_db = True
        else:
            raise ValueError(
                f"Path {self.db_path} does not exist and make_db is False"
            )
        
        try:
            self.engine = db.create_engine(
                f"sqlite:///{db_path}", echo=echo, **db_args
            )
            Session.configure(bind=self.engine)
            self.Session = sessionmaker(bind=self.engine)
            Base.metadata.create_all(self.engine)
        except db.exc.OperationalError as e:
            logger.error(f"Unable to connect to database file {db_path}")
            raise e

        # Defines frame_types
        if new_db:
            self._create_frame_types()

    def _create_frame_types(self):
        session = self.Session()
        if not session.query(FrameType).all():
            logger.info("Creating FrameType table...")
            for k in type_key:
                ft = FrameType(type_name=k)
                session.add(ft)
            session.commit()

    def _start_finalization(self, session):
        """Initialize finalization rows if they are not in the database yet"""
        session = self.Session()
        row = session.query(Finalize).filter(Finalize.agent == "G3tSMURF").one_or_none()
        if row is None:
            session.add(Finalize(agent="G3tSMURF", time=1.6e9))
            session.commit()
        for server in self.finalize.get("servers", []):
            for x in ["smurf-suprsync", "timestream-suprsync"]:
                agent = server.get(x)
                if agent is None:
                    continue
                row = (
                    session.query(Finalize)
                    .filter(Finalize.agent == agent)
                    .one_or_none()
                )
                if row is not None:
                    continue
                session.add(Finalize(agent=agent, time=1.6e9))
                session.commit()

    @classmethod
    def from_configs(cls, configs, **kwargs):
        """
        Create a G3tSmurf instance from a configs dictionary or yaml file
        example configuration file will all relevant entries::


            data_prefix : "/path/to/daq-node/"
            g3tsmurf_db: "/path/to/g3tsmurf.db"
            g3thk_db: "/path/to/g3hk.db"

            finalization:
                servers:
                    - smurf-suprsync: "smurf-sync-so1" ## instance-id
                      timestream-suprsync: "timestream-sync-so1" ## instance-id
                      pysmurf-monitor: "monitor-so1" ## instance-id
                    - smurf-suprsync: "smurf-sync-so2" ## instance-id
                      timestream-suprsync: "timestream-sync-so2" ## instance-id
                      pysmurf-monitor: "monitor-so2" ## instance-id

        The HK and finalization entries in the configuration are required for
        maintaining the g3tsmurf database but not for simply accessing or
        querying information.

        Args
        -----
        configs - dictionary containing `data_prefix` and `g3tsmurf_db` keys
        """
        if isinstance(configs, str):
            configs = load_configs(configs)

        return cls(
            os.path.join(configs["data_prefix"], "timestreams"),
            configs["g3tsmurf_db"],
            meta_path=os.path.join(configs["data_prefix"], "smurf"),
            db_args=configs.get("db_args", {}),
            finalize=configs.get("finalization", {}),
            hk_db_path=configs.get("g3thk_db"),
            **kwargs
        )

    @staticmethod
    def _make_datetime(x):
        """
        Takes an input (either a timestamp or datetime), and returns a datetime.
        Intended to allow flexibility in inputs for various other functions
        Note that x will be assumed to be in UTC if timezone is not specified

        Args
        ----
            x: input datetime of timestamp

        Returns
        ----
            datetime: datetime of x if x is a timestamp
        """
        if np.issubdtype(type(x), np.floating) or np.issubdtype(type(x), np.integer):
            return dt.datetime.utcfromtimestamp(x)
        elif isinstance(x, np.datetime64):
            return x.astype(dt.datetime).replace(tzinfo=dt.timezone.utc)
        elif isinstance(x, dt.datetime) or isinstance(x, dt.date):
            if x.tzinfo == None:
                return x.replace(tzinfo=dt.timezone.utc)
            return x
        raise (Exception("Input not a datetime or timestamp"))

    @property
    def last_db_update(self):
        with self.Session() as session:
            t = session.query(Finalize).filter(Finalize.agent == "G3tSMURF").one().time
        return t

    @last_db_update.setter
    def last_update(self, time):
        with self.Session() as session:
            agent = session.query(Finalize).filter(Finalize.agent == "G3tSMURF").one()
            agent.time = time
            session.commit()

    def add_file(self, path, session, overwrite=False):
        """
        Indexes a single file and adds it to the sqlite database. Creates a
        single entry in Files and as many Frame entries as there are frames in
        the file.

        Args
        ----
            path: path
                Path of the file to index
            session : SQLAlchemy session
                Current, active sqlalchemy session
            overwrite : bool
                If true and file exists in the database, update it.
        """

        frame_types = {ft.type_name: ft for ft in session.query(FrameType).all()}

        # name has a unique constraint in table
        db_file = session.query(Files).filter(Files.name == path).one_or_none()
        if db_file is None:
            db_file = Files(name=path)
            session.add(db_file)
        elif not overwrite:
            logger.info(f"File {path} found in database, use overwrite=True to update")
            return
        else:
            logger.debug(
                f"File {path} found in database, updating entry and re-making frames"
            )
            db_frames = db_file.frames
            [session.delete(frame) for frame in db_frames]
            session.commit()

        status = SmurfStatus.from_file(db_file.name)

        if status.tune is None or status.tune == "":
            my_tune = None
        else:
            my_tune = (
                session.query(Tunes)
                .filter(Tunes.name == Tunes.get_name_from_status(status))
                .one_or_none()
            )
        if my_tune is not None:
            ## tune has been found in streamed data.
            self.add_tuneset_to_file(my_tune, db_file, session)

        db_file.stream_id = status.stream_id
        if db_file.stream_id is None:
            try:
                splits = path.split("/")
                db_file.stream_id = splits[-2]
            except:
                logger.warning(f"Failed to find stream_id for {path}")
                pass

        reader = so3g.G3IndexedReader(path)

        total_channels = 0
        file_start, file_stop = None, None
        frame_idx = -1
        timing = None

        while True:
            try:
                db_frame_offset = reader.Tell()
                frames = reader.Process(None)
                if not frames:
                    break
            except RuntimeError as e:
                logger.warning(f"Failed to add {path}: file likely corrupted")
                session.rollback()
                return

            frame = frames[0]
            frame_idx += 1

            if str(frame.type) not in type_key:
                continue

            db_frame_frame_type = frame_types[str(frame.type)]

            timestamp = frame["time"].time / spt3g_core.G3Units.s
            db_frame_time = dt.datetime.utcfromtimestamp(timestamp)

            if str(frame.type) != "Wiring":
                dump = False
            else:
                dump = bool(frame["dump"])

            # only make Frame once the non-nullable fields are known
            db_frame = Frames(
                frame_idx=frame_idx,
                file=db_file,
                offset=db_frame_offset,
                frame_type=db_frame_frame_type,
                time=db_frame_time,
                status_dump=dump,
            )

            data = frame.get("data")
            sostream_version = frame.get("sostream_version", 0)
            if data is not None:
                if sostream_version >= 2:  # Using SuperTimestreams
                    db_frame.n_channels = len(data.names)
                    db_frame.n_samples = len(data.times)
                    db_frame.start = dt.datetime.utcfromtimestamp(
                        data.times[0].time / spt3g_core.G3Units.s
                    )
                    db_frame.stop = dt.datetime.utcfromtimestamp(
                        data.times[-1].time / spt3g_core.G3Units.s
                    )
                    if timing is None:
                        timing = frame.get("timing_paradigm", "") == "High Precision"
                    else:
                        timing = timing and (
                            frame.get("timing_paradigm", "") == "High Precision"
                        )
                    fo = frame.get('primary', None)
                    if fo is None:
                        timing = False # no good timing without primary
                    else:
                        key_map = {k: i for i, k in enumerate(fo.names)}
                        counters = np.all(
                            np.diff( fo.data[ key_map['Counter0']] ) != 0 
                        ) and np.all( 
                            np.diff( fo.data[ key_map['Counter2']] ) != 0
                        )
                        # check all counters are incrementing
                        timing = timing and counters 

                else:
                    db_frame.n_samples = data.n_samples
                    db_frame.n_channels = len(data)
                    db_frame.start = dt.datetime.utcfromtimestamp(
                        data.start.time / spt3g_core.G3Units.s
                    )
                    db_frame.stop = dt.datetime.utcfromtimestamp(
                        data.stop.time / spt3g_core.G3Units.s
                    )

                if file_start is None:
                    file_start = db_frame.start
                file_stop = db_frame.stop
                total_channels = max(total_channels, db_frame.n_channels)

            session.add(db_frame)

        if file_start is None:
            ## happens if there are no scan frames in file
            file_start = db_file.frames[0].time
            file_stop = db_file.frames[-1].time

        db_file.start = file_start
        db_file.stop = file_stop
        db_file.n_channels = total_channels
        db_file.n_frames = frame_idx
        db_file.timing = timing
        session.commit()

        if len(status.tags) > 0:
            if status.tags[0] == "obs" or status.tags[0] == "oper":
                ## this is where I tell it to make an observation
                logger.debug(f"file {db_file.name} is an observation")
                self.add_new_observation_from_status(status, session)

    def index_archive(
        self,
        stop_at_error=False,
        skip_old_format=True,
        min_ctime=None,
        max_ctime=None,
        show_pb=True,
        session=None,
    ):
        """
        Adds all files from an archive to the File and Frame sqlite tables.
        Files must be indexed before the metadata entries can be made.

        Args
        ----
        stop_at_error: bool
            If True, will stop if there is an error indexing a file.
        skip_old_format: bool
            If True, will skip over indexing files before the name convention
            was changed to be ctime_###.g3.
        min_ctime: int, float, or None
            If set, files with session-ids less than this ctime will be
            skipped.
        max_ctime: int, float, or None
            If set, files with session-ids higher than this ctime will be
            skipped.
        show_pb: bool
            If true, will show progress bar for file indexing
        """
        new_session = False
        if session is None:
            session = self.Session()
            new_session=True
        indexed_files = [f[0] for f in session.query(Files.name).all()]

        files = []
        for root, _, fs in os.walk(self.archive_path):
            for f in fs:
                path = os.path.join(root, f)
                if path.endswith(".g3") and path not in indexed_files:
                    if skip_old_format and "2020-" in path:
                        continue

                    if "-" not in f and (min_ctime is not None):
                        # We know the filename is <ctime>_###.g3
                        session_id = int(f.split("_")[0])
                        if session_id < min_ctime:
                            continue
                    if "-" not in f and (max_ctime is not None):
                        # We know the filename is <ctime>_###.g3
                        session_id = int(f.split("_")[0])
                        if session_id > max_ctime:
                            continue
                    files.append(path)

        logger.info(f"Indexing {len(files)} files...")

        ## files must be updated in sequencial order. otherwise we may end up
        ## with more TuneSets than are necessary
        for f in tqdm(sorted(files), disable=(not show_pb)):
            try:
                self.add_file(os.path.join(root, f), session)
                session.commit()
            except IntegrityError as e:
                # Database Integrity Errors, such as duplicate entries
                session.rollback()
                logger.warning(f"Integrity error with error {e}")
            except RuntimeError as e:
                # End of stream errors, for G3Files that were not fully flushed
                session.rollback()
                logger.warning(f"Failed on file {f} due to end of stream error!")
            except Exception as e:
                # This will catch generic errors such as attempting to load
                # out-of-date files that do not have the required frame
                # structure specified in the TOD2MAPS docs.
                session.rollback()
                if stop_at_error:
                    raise e
                logger.warning(f"Failed on file {f}:\n{e}")

        if new_session:
            session.close()

    def delete_file(self, db_file, session=None, dry_run=False, my_logger=None):
        """WARNING: Deletes data from the file system

        Delete both a database file entry, it's associated frames, AND the
        file itself. Only to be run by automated data management systems

        Args
        ----
        db_file: File instance
            database Fine instance to be deleted
        session: optional, SQLAlchemy session
            should be passed if file is called as part of a larger cleanup
            function
        dry_run: boolean
            if true, just prints deletion to my_logger.info
        my_logger: logger, optional
            option to pass different logger to this function
        """

        if session is None:
            session = self.Session()
        if my_logger is None:
            my_logger = logger

        db_frames = db_file.frames
        my_logger.debug(f"Deleting frame entries for {db_file.name}")
        if not dry_run:
            [session.delete(frame) for frame in db_frames]

        if not os.path.exists(db_file.name):
            my_logger.warning(
                f"Database file {db_file.name} appears already deleted on disk"
            )
        else:
            my_logger.debug(f"Deleting file {db_file.name}")

            if not dry_run:
                os.remove(db_file.name)

                ## clean up directory if it is empty
                base, _ = os.path.split(db_file.name)
                if len(os.listdir(base)) == 0:
                    os.rmdir(base)
        my_logger.debug(f"Deleting database entry for {db_file.name}")
        if not dry_run:
            session.delete(db_file)
            session.commit()

    def add_new_channel_assignment(self, stream_id, ctime, cha, cha_path, session):
        """Add new entry to the Channel Assignments table. Called by the
        index_metadata function.

        Args
        -------
        stream_id : string
            The stream id for the particular SMuRF slot
        ctime : int
            The ctime of the SMuRF action called to create the channel
            assignemnt
        cha : string
            The file name of the channel assignment
        cha_path : path
            The absolute path to the channel assignment
        session : SQLAlchemy Session
            The active session
        """

        band = int(re.findall("b\d.txt", cha)[0][1])

        ch_assign = session.query(ChanAssignments).filter(
            ChanAssignments.ctime == ctime,
            ChanAssignments.stream_id == stream_id,
            ChanAssignments.band == band,
        )
        ch_assign = ch_assign.one_or_none()
        if ch_assign is None:
            ch_assign = ChanAssignments(
                ctime=ctime, path=cha_path, name=cha, stream_id=stream_id, band=band
            )
            session.add(ch_assign)
            session.commit()

    def add_channels_from_assignment(self, ch_assign, session):
        """Add the channels that are associated with a particular channel
        assignment entry

        Args
        ----
        ch_assign: ChanAssignment instance
        session: session used to find ch_assign
        """

        notches = np.atleast_2d(np.genfromtxt(ch_assign.path, delimiter=","))
        if np.sum([notches[:, 2] != -1]) != len(ch_assign.channels):
            ch_made = [c.channel for c in ch_assign.channels]
            for notch in notches:
                # smurf did not assign a channel
                if notch[2] == -1:
                    continue
                if int(notch[2]) in ch_made:
                    continue

                ch_name = "sch_{}_{:10d}_{:01d}_{:03d}".format(
                    ch_assign.stream_id, ch_assign.ctime, ch_assign.band, int(notch[2])
                )
                ch = Channels(
                    subband=int(notch[1]),
                    channel=int(notch[2]),
                    frequency=notch[0],
                    name=ch_name,
                    chan_assignment=ch_assign,
                    band=ch_assign.band,
                )

                if ch.channel == -1:
                    logger.warning(
                        f"Un-assigned channel made in Channel Assignment {ch_assign.name}"
                    )
                    continue
                check = (
                    session.query(Channels)
                    .filter(
                        Channels.ca_id == ch_assign.id, Channels.channel == ch.channel
                    )
                    .one_or_none()
                )
                if check is None:
                    session.add(ch)
        session.commit()

    def _assign_set_from_file(
        self, tune_path, ctime=None, stream_id=None, session=None
    ):
        """Build set of Channel Assignments that are (or should be) in the
        tune file.

        Args
        -------
        tune_path : path
            The absolute path to the tune file
        ctime : int
            ctime of SMuRF Action
        stream_id : string
            The stream id for the particular SMuRF slot
        session : SQLAlchemy Session
            The active session
        """

        if session is None:
            session = self.Session()
        if stream_id is None:
            stream_id = tune_path.split("/")[-4]
        if ctime is None:
            ctime = int(tune_path.split("/")[-1].split("_")[0])

        data = np.load(tune_path, allow_pickle=True).item()
        assign_set = []

        # Determine the TuneSet
        for band in data.keys():
            if "resonances" not in data[band]:
                # tune file doesn't have info for this band
                continue
            # try to use tune file to find channel assignment before just
            # assuming "most recent"
            if "channel_assignment" in data[band]:
                cha_name = data[band]["channel_assignment"].split("/")[-1]
                cha = (
                    session.query(ChanAssignments)
                    .filter(
                        ChanAssignments.stream_id == stream_id,
                        ChanAssignments.name == cha_name,
                    )
                    .one_or_none()
                )
            else:
                cha = session.query(ChanAssignments).filter(
                    ChanAssignments.stream_id == stream_id,
                    ChanAssignments.ctime <= ctime,
                    ChanAssignments.band == band,
                )

                cha = cha.order_by(db.desc(ChanAssignments.ctime)).first()
            if cha is None:
                logger.error(f"Missing Channel Assignment for tune file {tune_path}")
                continue
            assign_set.append(cha)
        return assign_set

    def add_new_tuning(self, stream_id, ctime, tune_path, session):
        """Add new entry to the Tune table, Called by the index_metadata function.

        Args
        -------
        stream_id : string
            The stream id for the particular SMuRF slot
        ctime : int
            The ctime of the SMuRF action called to create the tuning file.
        tune_path : path
            The absolute path to the tune file
        session : SQLAlchemy Session
            The active session
        """
        name = stream_id + "_" + tune_path.split("/")[-1].strip(".npy")
        tune = (
            session.query(Tunes)
            .filter(Tunes.name == name, Tunes.stream_id == stream_id)
            .one_or_none()
        )
        if tune is None:
            tune = Tunes(
                name=name,
                start=dt.datetime.utcfromtimestamp(ctime),
                path=tune_path,
                stream_id=stream_id,
            )
            session.add(tune)
            session.commit()

    def add_tuneset_to_file(self, db_tune, db_file, session):
        """Uses tune found in file to decide we should create a tuneset. Assigns
        both tune and tuneset to the file entry. called by add_file

        Args
        ----
        db_tune: Tunes instance for tune in file
        db_file: File instance for file
        session: sqlalchemy session used to create instances
        """

        if db_tune.tuneset is None:
            # assign set is the set of channel assignments that make up this tune file
            assign_set = self._assign_set_from_file(
                db_tune.path, stream_id=db_tune.stream_id, session=session
            )

            tuneset = None
            tunesets = session.query(TuneSets)
            # tunesets with any channel assignments matching list
            tunesets = tunesets.filter(
                TuneSets.chan_assignments.any(
                    ChanAssignments.id.in_([a.id for a in assign_set])
                )
            ).all()

            if len(tunesets) > 0:
                for ts in tunesets:
                    if np.all(
                        sorted([ca.id for ca in ts.chan_assignments])
                        == sorted([a.id for a in assign_set])
                    ):
                        tuneset = ts

            if tuneset is None:
                logger.debug(
                    f"New Tuneset Detected {db_tune.name},"
                    f" {[[a.name for a in assign_set]]}"
                )
                tuneset = TuneSets(
                    name=db_tune.name,
                    path=db_tune.path,
                    stream_id=db_tune.stream_id,
                    start=db_tune.start,
                )
                session.add(tuneset)
                session.commit()

                # add a the assignments and channels to the detector set
                for db_cha in assign_set:
                    ## make channels now that the channel assignment is part
                    ## of a tuneset
                    self.add_channels_from_assignment(db_cha, session=session)
                    tuneset.chan_assignments.append(db_cha)
                    for ch in db_cha.channels:
                        tuneset.channels.append(ch)
                session.commit()

            db_tune.tuneset = tuneset
            session.commit()

        db_file.tune = db_tune
        db_file.tuneset = db_tune.tuneset
        session.commit()

    def add_new_observation_from_status(self, status, session):
        """Wrapper to pull required information from SmurfStatus and create a
        new observation. Works based on tags in Status frame, so may not work
        with older files.

        Args
        ----
        status : SmurfStatus instance
        session : SQLAlchemy Session
            The active session
        """

        self.add_new_observation(
            status.stream_id,
            status.action,
            status.action_timestamp,
            session,
            calibration=(status.tags[0] == "oper"),
            session_id=status.session_id,
        )

    def add_new_observation(
        self,
        stream_id,
        action_name,
        action_ctime,
        session,
        calibration,
        session_id=None,
        max_early=5,
    ):
        """Add new entry to the observation table. Called by the
        index_metadata function.

        Args
        -------
        stream_id : string
            The stream id for the particular SMuRF slot
        action_ctime : int
            The ctime of the SMuRF action called to create the observation. Often
            slightly different than the .g3 session ID
        session : SQLAlchemy Session
            The active session
        calibration : boolean
            Boolean that indicates whether the observation is a calibration observation.
        session_id : int (optional, but much more efficient)
            session id, if known, for timestream files that should go with the observations
        max_early : int (optional)
            Buffer time to allow the g3 file to be earlier than the smurf action
        """

        if session_id is None:
            db_file = session.query(Files)
            db_file = db_file.filter(
                Files.stream_id == stream_id,
                Files.start >= dt.datetime.utcfromtimestamp(action_ctime - max_early),
            )
            db_file = db_file.order_by(Files.start).first()
            if db_file is None:
                logger.debug(
                    f"No .g3 files from Action {action_name} in {stream_id}"
                    f" at {action_ctime}. Not Making Observation"
                )
                return

            session_id = int((db_file.name[:-3].split("/")[-1]).split("_")[0])

        if calibration:
            obs_id = f"oper_{stream_id}_{session_id}"
        else:
            obs_id = f"obs_{stream_id}_{session_id}"

        # Check if observation exists already
        obs = (
            session.query(Observations)
            .filter(
                Observations.obs_id == obs_id,
                Observations.stream_id == stream_id,
                Observations.action_name == action_name,
                Observations.action_ctime == action_ctime,
            )
            .one_or_none()
        )
        logger.debug(f"Observations {obs_id} already exists")

        if obs is None:
            db_file = session.query(Files)
            db_file = db_file.filter(
                Files.stream_id == stream_id, Files.name.like(f"%{session_id}%")
            )
            db_file = db_file.order_by(Files.start).first()
            logger.debug(f"Found file {db_file.name} to be the start of observation {obs_id}")
            if db_file.obs_id is not None:
                if db_file.obs_id != obs_id:
                    logger.warning(f"Trying to make {obs_id} using file {db_file.name} "
                                   f"but file is already part if {db_file.obs_id}. Will "
                                   f"not create observation")
                    return
            
            # Verify the files we found match with Observation
            status = SmurfStatus.from_file(db_file.name)
            if status.action is not None:
                assert status.action == action_name
                assert status.action_timestamp == action_ctime

            # Verify inside of file matches the outside
            reader = so3g.G3IndexedReader(db_file.name)
            while True:
                frames = reader.Process(None)
                if not frames:
                    break
                frame = frames[0]

                if str(frame.type) == "Observation":
                    assert frame["sostream_id"] == stream_id
                    assert frame["session_id"] == session_id
                    break

            # Build Observation
            obs = Observations(
                obs_id=obs_id,
                timestamp=session_id,
                action_ctime=action_ctime,
                action_name=action_name,
                stream_id=stream_id,
                calibration=calibration,
            )
            session.add(obs)
            session.commit()

        # obs.stop is only updated when streaming session is over
        if obs.stop is None:
            self.update_observation_files(
                obs,
                session,
                max_early=max_early,
            )

    def update_observation_files(
        self, 
        obs, 
        session, 
        max_early=5, 
        force=False, 
        force_stop=False,
    ):
        """Update existing observation. A separate function to make it easier
        to deal with partial data transfers. See add_new_observation for args

        Args
        -----
        max_early : int
            Buffer time to allow the g3 file to be earlier than the smurf action
        session : SQLAlchemy Session
            The active session
        force : bool
            If true, will recalculate file/tune information even if observation
            appears complete
        force_stop: bool
            If true, will force the end of the observation to be set to the end
            of the current file list. Useful for completing observations where
            computer systems crashed/restarted during the observations so no end
            frames were written.
        """

        if not force and obs.stop is not None:
            logger.debug(f"Returning from {obs.obs_id} without updates")
            return

        # Prefix is deterministic based on observation details
        prefix = os.path.join(self.archive_path, str(obs.timestamp)[:5], obs.stream_id)

        flist = (
            session.query(Files)
            .filter(Files.name.like(prefix + "/" + str(obs.timestamp) + "%"))
            .order_by(Files.start)
            .all()
        )

        logger.debug(f"Found {len(flist)} files in {obs.obs_id}")
        if len(flist) == 0:
            ## we don't have .g3 files for some reason, shouldn't be possible?
            logger.debug(f"Found no files associated with {obs.obs_id}")
            return

        ## set start to be the first scan frame in file/observation
        obs.start = flist[0].start

        ## Load Status Information
        status = SmurfStatus.from_file(flist[0].name)

        # Add any tags from the status
        if len(status.tags) > 0:
            for tag in status.tags:
                t = (
                    session.query(Tags)
                    .filter(Tags.tag == tag, Tags.obs_id == obs.obs_id)
                    .one_or_none()
                )
                if t is not None:
                    continue
                new_tag = Tags(obs_id=obs.obs_id, tag=tag)
                session.add(new_tag)
            obs.tag = ",".join(status.tags)

        # Add Tune and Tuneset information
        if status.tune is not None and status.tune != "":
            tune = (
                session.query(Tunes)
                .filter(
                    Tunes.name == Tunes.get_name_from_status(status),
                    Tunes.stream_id == obs.stream_id,
                )
                .one_or_none()
            )
            if tune is None:
                logger.warning(
                    f"Tune {status.tune} not found in database, update error?"
                )
                tuneset = None
            else:
                tuneset = tune.tuneset
        else:
            tune = None
            tuneset = session.query(TuneSets).filter(
                TuneSets.start <= obs.start,
                TuneSets.stream_id == obs.stream_id,
            )
            tuneset = tuneset.order_by(db.desc(TuneSets.start)).first()

        already_have = [ts.id for ts in obs.tunesets]
        if tuneset is not None:
            if not tuneset.id in already_have:
                obs.tunesets.append(tuneset)

        obs_samps = 0
        # Update file entries
        for db_file in flist:
            db_file.obs_id = obs.obs_id
            if db_file.tuneset is not None and tuneset is not None:
                db_file.detset = tuneset.name
            if db_file.tune is not None and tune is not None:
                db_file.tune_id = tune.id

            # this is where I learned sqlite does not accept numpy 32 or 64 bit ints
            file_samps = sum(
                [
                    fr.n_samples if fr.n_samples is not None else 0
                    for fr in db_file.frames
                ]
            )
            db_file.sample_start = obs_samps
            db_file.sample_stop = obs_samps + file_samps
            obs_samps = obs_samps + file_samps

        obs_ended = _file_has_end_frames(flist[-1].name)
        if (
            obs.stop is None
            and not obs_ended
            and (force or flist[-1].stop <= dt.datetime.now() - dt.timedelta(hours=1))
        ):
            ## try to brute force find the end if it's been awhile
            for f in flist[::-1][:3]:
                if not obs_ended:
                    obs_ended = _file_has_end_frames(f.name)

        if obs_ended or force_stop:
            logger.debug(f"Found that {obs.obs_id} has ended.")
            obs.n_samples = obs_samps
            obs.duration = flist[-1].stop.timestamp() - flist[0].start.timestamp()
            obs.stop = flist[-1].stop
            # f.timing is None if the file has no Scan Frames
            # timing requires external downsampling AND high precision 
            # timestamps. file.timing checks high precision. obs.timing
            # includes external downsampling
            obs.timing = (
                np.all([f.timing or (f.timing is None) for f in flist]) 
                and status.downsample_external
            )
            ## data not usable for science
            if not obs.timing:
                logger.warning(f"{obs.obs_id} has bad timing. Updating tags")
                tags = obs.tag.split(',')
                tags.insert(1, 'bad')
                tags[2] = f'bad_{tags[2]}'
                tags.append('timing_issues')
                obs.tag = ','.join(tags)
            logger.debug(f"Setting {obs.obs_id} stop time to {obs.stop}")
        session.commit()

    def delete_observation_files(
        self, obs, session, dry_run=False, my_logger=None
    ):
        """WARNING: Deletes files from the file system

        Args
        ----
        obs: observation instance
        session: SQLAlchemy session used to query obs
        dry_run: boolean
            if true, only prints deletion to my_logger.info
        """
        if my_logger is None:
            my_logger = logger

        my_logger.info(f"Deleting Observation {obs.obs_id}")
        ## first remove the tags
        tags = (
            session.query(Tags)
            .filter(
                Tags.obs_id == obs.obs_id,
            )
            .all()
        )
        for t in tags:
            my_logger.debug(f"Deleting Tag ({t.tag, t.obs_id}) from database")
            if not dry_run:
                session.delete(t)

        ## then remove the files
        for f in obs.files:
            self.delete_file(f, session, dry_run=dry_run, my_logger=my_logger)

        ## then remove the observation
        my_logger.debug(f"Deleting Observation {obs.obs_id} from database")
        if not dry_run:
            session.delete(obs)
            session.commit()

    def search_metadata_actions(
        self, min_ctime=16000 * 1e5, max_ctime=None, reverse=False
    ):
        """Generator used to page through smurf folder returning each action
        formatted for easy use.

        Args
        -----
        min_ctime : lowest timestamped action to return
        max_ctime : highest timestamped action to return
        reverse : if true, goes backward

        Yields
        -------
        tuple (action, stream_id, ctime, path)
        action : Smurf Action string with ctime removed for easy comparison
        stream_id : stream_id of Action
        ctime : ctime of Action folder
        path : absolute path to action folder
        """
        if max_ctime is None:
            max_ctime = dt.datetime.now().timestamp()

        if self.meta_path is None:
            raise ValueError(
                "Archiver needs meta_path attribute to index channel assignments"
            )

        logger.debug(f"Ignoring ctime folders below {int(min_ctime//1e5)}")

        for ct_dir in sorted(os.listdir(self.meta_path), reverse=reverse):
            if int(ct_dir) < int(min_ctime // 1e5):
                continue
            elif int(ct_dir) > int(max_ctime // 1e5):
                continue

            for stream_id in sorted(
                os.listdir(os.path.join(self.meta_path, ct_dir)), reverse=reverse
            ):
                action_path = os.path.join(self.meta_path, ct_dir, stream_id)
                actions = sorted(os.listdir(action_path), reverse=reverse)

                for action in actions:
                    try:
                        ctime = int(action.split("_")[0])
                        if ctime < min_ctime or ctime > max_ctime:
                            continue
                        astring = "_".join(action.split("_")[1:])

                        yield (
                            astring,
                            stream_id,
                            ctime,
                            os.path.join(action_path, action),
                        )
                    except GeneratorExit:
                        return
                    except:
                        continue

    def search_metadata_files(
        self,
        min_ctime=16000 * 1e5,
        max_ctime=None,
        reverse=False,
        skip_plots=True,
        skip_configs=True,
    ):
        """Generator used to page through smurf folder returning each file
        formatted for easy use.

        Args
        -----
        min_ctime : int or float
            Lowest timestamped action to return
        max_ctime : int or float
            highest timestamped action to return
        reverse   : bool
            if true, goes backward
        skip_plots : bool
            if true, skips all the plots folders because we probably don't want
            to look through them
        skip_configs : bool
            if true, skips all the config folders because we probably don't want
            to look through them

        Yields
        -------
        tuple (fname, stream_id, ctime, abs_path)
        fname : string
            file name with ctime removed
        stream_id : string
            stream_id where the file is saved
        ctime : int
            file ctime
        abs_path : string
            absolute path to file
        """
        for action, stream_id, actime, path in self.search_metadata_actions(
            min_ctime=min_ctime, max_ctime=max_ctime, reverse=reverse
        ):
            if skip_configs and action == "config":
                continue
            adirs = os.listdir(path)
            for adir in adirs:
                if skip_plots and adir == "plots":
                    continue
                for root, dirs, files in os.walk(
                    os.path.join(path, adir), topdown=False
                ):
                    for name in files:
                        try:
                            try:
                                ctime = int(name.split("_")[0])
                            except ValueError:
                                ctime = actime
                            fname = "_".join(name.split("_")[1:])
                            yield (fname, stream_id, ctime, os.path.join(root, name))
                        except GeneratorExit:
                            return

    def search_suprsync_files(
        self, min_ctime=16000 * 1e5, max_ctime=None, reverse=False
    ):
        """Generator used to page through smurf folder returning each suprsync
        finalization file formatted for easy use

        Args
        -----
        min_ctime : lowest timestamped action to return
        max_ctime : highest timestamped action to return
        reverse : if true, goes backward

        Yields
        -------
        tuple (action, stream_id, ctime, path)
        action : Smurf Action string with ctime removed for easy comparison
        stream_id : stream_id of Action
        ctime : ctime of Action folder
        path : absolute path to action folder
        """
        if max_ctime is None:
            max_ctime = dt.datetime.now().timestamp()

        logger.debug(f"Ignoring ctime folders below {int(min_ctime//1e5)}")

        for base_dir in [self.archive_path, self.meta_path]:
            for ct_dir in sorted(os.listdir(base_dir), reverse=reverse):
                if int(ct_dir) < int(min_ctime // 1e5):
                    continue
                elif int(ct_dir) > int(max_ctime // 1e5):
                    continue

                if "suprsync" not in os.listdir(os.path.join(base_dir, ct_dir)):
                    continue
                base = os.path.join(base_dir, ct_dir, "suprsync")

                for agent in os.listdir(base):
                    for f in os.listdir(os.path.join(base, agent)):
                        try:
                            yield os.path.join(base, agent, f)
                        except GeneratorExit:
                            return

    def index_timecodes(self, session=None, min_ctime=16000e5, max_ctime=None):
        """Index the timecode finalizeation files coming out of suprsync"""
        if session is None:
            session = self.Session()

        if max_ctime is None:
            max_ctime = dt.datetime.now().timestamp()

        for fname in self.search_suprsync_files(
            min_ctime=min_ctime, max_ctime=max_ctime
        ):
            info = yaml.safe_load(open(fname, "r"))

            for subdir in info["subdirs"]:
                if subdir == "suprsync":
                    continue

                tcf = (
                    session.query(TimeCodes)
                    .filter(
                        TimeCodes.stream_id == subdir,
                        TimeCodes.suprsync_type == SupRsyncType.from_string(
                            info["archive_name"]
                        ).value,
                        TimeCodes.timecode == info["timecode"],
                    )
                    .one_or_none()
                )
                if tcf is not None:
                    continue
                tcf = TimeCodes(
                    stream_id=subdir,
                    suprsync_type=SupRsyncType.from_string(info["archive_name"]).value,
                    timecode=info["timecode"],
                    agent=info["instance_id"],
                )
                session.add(tcf)
                session.commit()

    def get_HK(self):
        if self.hk_db_path is None:
            raise ValueError("HK database path required")

        if self.HK is None:
            iids = []
            for server in self.finalize.get("servers", []):
                for key in server.keys():
                    # Append the value (iid) to the iids list
                    iids.append(server[key])

            self.HK = G3tHk(
                os.path.join(os.path.split(self.archive_path)[0], "hk"),
                iids = iids,
                db_path = self.hk_db_path,
            )
        return self.HK

    def update_finalization(self, update_time, session=None):
        """Update the finalization time rows in the database"""
        
        if session is None:
            session = self.Session()
        # look for new rows to add to table
        self._start_finalization(session)

        HK = self.get_HK()

        agent_list = session.query(Finalize).all()
        for agent in agent_list:
            if agent.agent == "G3tSMURF":
                continue
            db_agent = (
                HK.session.query(HKAgents)
                .filter(
                    HKAgents.instance_id == agent.agent,
                    HKAgents.start <= update_time,
                )
                .order_by(db.desc(HKAgents.start))
                .first()
            )
            if db_agent is None:
                logger.info(
                    f"Agent {agent.agent} not found in HK database before"
                    f" update time {update_time}"
                )
                continue
            f = [f for f in db_agent.fields if "finalized_until" in f.field]
            if len(f) == 0:
                logger.warning(
                    f"Did not find finalized_until in Agent {agent}"
                    f"from file {db_agent.hkfile.filename}"
                )
                continue
            f = f[0]
            data = HK.load_data(f)
            x = np.where(data[f.field][0] <= update_time)[0]
            if len(x) < 1:
                logger.error(
                    f"No data points before update time for agent "
                    f"{agent} in file {db_agent.hkfile.filename}?"
                )
            x = x[0]
            agent.time = data[f.field][1][x]

        session.commit()

    def get_final_time(
        self, stream_ids, start=None, stop=None, check_control=True,
        session=None
    ):
        """Return the ctime to which database is finalized for a set of 
        stream_ids between ctimes start and stop. If check_control is True it 
        will use the pysmurf-monitor entries in the HK database to determine 
        which pysmurf-monitors were in control of which stream_ids between 
        start and stop.
        """
        if check_control and self.hk_db_path is None:
            raise ValueError(
                "HK database path required to update finalization time"
            )
        if check_control and ((start is None) or (stop is None)):
            raise ValueError(
                "start and stop ctimes are required to check which"
                " pysmurf-monitors control which stream_ids"
            )
        if session is None:
            session = self.Session()
        
        HK = self.get_HK()

        agent_list = []
        if "servers" not in self.finalize:
            raise ValueError(
                "server list not found in finalization "
                "information, required to check completeness"
            )
        for server in self.finalize.get("servers"):
            if "smurf-suprsync" not in server:
                logger.error(
                    f"Incomplete finalization information, missing "
                    f"'smurf-suprync' in {server}"
                )
            if "timestream-suprsync" not in server:
                logger.error(
                    f"Incomplete finalization information, missing "
                    f"'timestream-suprync' in {server}"
                )
            pm = server.get("pysmurf-monitor")
            if not check_control or pm is None:
                if check_control:
                    logger.warning(
                        f"Incomplete finalization information, "
                        f"missing 'pysmurf-monitor'. Not checking "
                        "for stream_id control"
                    )
                agent_list.append(server["smurf-suprsync"])
                agent_list.append(server["timestream-suprsync"])
                continue
            if stop > HK.get_last_update():
                if stop > HK.get_last_update()+3600:
                    logger.error(f"HK database not updated recently enough to"
                                  " check finalization time. Last update "
                                 f"{HK.get_last_update}. Trying to check until"
                                 f"{stop}")
                else:
                    stop = HK.get_last_update()
            sids = pysmurf_monitor_control_list(pm, start, stop, HK)
            if np.any([s in stream_ids for s in sids]):
                agent_list.append(server["smurf-suprsync"])
                agent_list.append(server["timestream-suprsync"])

        min_time = self.last_update
        for agent in agent_list:
            (time,) = session.query(Finalize.time).filter(Finalize.agent == agent).one()
            min_time = min(min_time, time)
        return min_time

    def _process_index_error(self, session, e, stream_id, ctime, path, stop_at_error):
        if type(e) == ValueError:
            logger.info(f"Value Error at {stream_id}, {ctime}, {path}")
        elif type(e) == IntegrityError:
            # Database Integrity Errors, such as duplicate entries
            session.rollback()
            logger.info(f"Integrity Error at {stream_id}, {ctime}, {path}")
        else:
            logger.info(
                f"Error of type {type(e).__name__} at {stream_id}, {ctime}, {path}"
            )
        if stop_at_error:
            raise (e)

    def index_channel_assignments(
        self,
        session,
        min_ctime=16000 * 1e5,
        max_ctime=None,
        pattern="channel_assignment",
        stop_at_error=False,
    ):
        """Index all channel assignments newer than a minimum ctime

        Args
        -----
        session : G3tSmurf session connection
        min_time : int of float
            minimum time for for indexing
        max_time : int, float, or None
            maximum time for indexing
        pattern : string
            string pattern to look for channel assignments
        """

        for fpattern, stream_id, ctime, path in self.search_metadata_files(
            min_ctime=min_ctime, max_ctime=max_ctime
        ):
            if pattern in fpattern:
                try:
                    # decide if this is the last channel assignment in the directory
                    # needed because we often get multiple channel assignments in the
                    # same folder
                    root = os.path.join("/", *path.split("/")[:-1])
                    fname = path.split("/")[-1]
                    fband = int(re.findall("b\d.txt", fname)[0][1])
                    cha_times = [
                        int(f.split("_")[0])
                        for f in os.listdir(root)
                        if f"b{fband}.txt" in f.split("_")[-1]
                    ]
                    if ctime != np.max(cha_times):
                        continue
                    logger.debug(
                        f"Add new channel assignment: {stream_id},{ctime}, {path}"
                    )
                    self.add_new_channel_assignment(
                        stream_id, ctime, fname, path, session
                    )
                except Exception as e:
                    self._process_index_error(
                        session, e, stream_id, ctime, path, stop_at_error
                    )

    def index_tunes(
        self,
        session,
        min_ctime=16000 * 1e5,
        max_ctime=None,
        pattern="tune.npy",
        stop_at_error=False,
    ):
        """Index all tune files newer than a minimum ctime

        Args
        -----
        session : G3tSmurf session connection
        min_time : int of float
            minimum time for indexing
        max_time : int, float, or None
            maximum time for indexing
        pattern : string
           string pattern to look for tune files
        """
        for fname, stream_id, ctime, path in self.search_metadata_files(
            min_ctime=min_ctime, max_ctime=max_ctime
        ):
            if pattern in fname:
                try:
                    logger.debug(f"Add new Tune: {stream_id}, {ctime}, {path}")
                    self.add_new_tuning(stream_id, ctime, path, session)
                except Exception as e:
                    self._process_index_error(
                        session, e, stream_id, ctime, path, stop_at_error
                    )

    def index_observations(
        self, session, min_ctime=16000 * 1e5, max_ctime=None, stop_at_error=False
    ):
        """Index all observations newer than a minimum ctime. Uses
        SMURF_ACTIONS to define which actions are observations.

        Args
        -----
        session : G3tSmurf session connection
        min_time : int or float
            minimum time for indexing
        max_time : int, float, or None
            maximum time for indexing
        """
        logger.warning("Indexing via actions is deprecated and SHOULD NOT be "
                       "run on systems where level 2 files are being automatically "
                       "deleted")
        for action, stream_id, ctime, path in self.search_metadata_actions(
            min_ctime=min_ctime, max_ctime=max_ctime
        ):
            if (
                action in SMURF_ACTIONS["observations"]
                or action in SMURF_ACTIONS["calibrations"]
            ):
                try:
                    obs_path = os.listdir(os.path.join(path, "outputs"))
                    logger.debug(
                        f"Add new Observation: {stream_id}, {ctime}, {obs_path}"
                    )
                    self.add_new_observation(
                        stream_id,
                        action,
                        ctime,
                        session,
                        calibration=(action in SMURF_ACTIONS["calibrations"]),
                    )
                except Exception as e:
                    self._process_index_error(
                        session, e, stream_id, ctime, path, stop_at_error
                    )

    def index_metadata(
        self, min_ctime=16000 * 1e5, max_ctime=None, stop_at_error=False,
        session=None,
    ):
        """Adds all channel assignments, tunes, and observations in archive to
        database. Adding relevant entries to Files as well.

        Args
        ----
        min_ctime : int
            Lowest ctime to start looking for new metadata
        max_ctime : None or int
            Highest ctime to look for new metadata
        stop_at_error: bool
           If True, will stop if there is an error indexing a file.
        """

        if self.meta_path is None:
            raise ValueError(
                "Archiver needs meta_path attribute to index channel assignments"
            )
        new_session = False
        if session is None:
            session = self.Session()
            new_session = True
        logger.debug(f"Ignoring ctime folders below {int(min_ctime//1e5)}")

        logger.debug("Indexing Channel Assignments")
        self.index_channel_assignments(
            session,
            min_ctime=min_ctime,
            max_ctime=max_ctime,
            stop_at_error=stop_at_error,
        )
        logger.debug("Indexing Tune Files")
        self.index_tunes(
            session,
            min_ctime=min_ctime,
            max_ctime=max_ctime,
            stop_at_error=stop_at_error,
        )
        if new_session:
            session.close()

    def index_action_observations(
        self, min_ctime=16000 * 1e5, max_ctime=None, stop_at_error=False,
        session=None,
    ):
        """Looks through Action folders to build Observations not built off of
        tags in add_file. This function is a hold-over from when tags were not
        used to find Observations (change made ~Jan. 2023)

        Args
        ----
        min_ctime : int
            Lowest ctime to start looking for new metadata
        max_ctime : None or int
            Highest ctime to look for new metadata
        stop_at_error: bool
           If True, will stop if there is an error indexing a file.
        """
        new_session = False
        if session is None:
            session = self.Session()
        logger.debug(f"Ignoring ctime folders below {int(min_ctime//1e5)}")

        logger.debug("Indexing Observations")
        self.index_observations(
            session,
            min_ctime=min_ctime,
            max_ctime=max_ctime,
            stop_at_error=stop_at_error,
        )
        if new_session:
            session.close()

    def find_missing_files(self, timecode, session=None):
        """create a list of files in the timecode folder that are not in the
        g3tsmurf database
        
        Arguments
        ----------
        timecode (int): a level 2 timestreams timecode

        Returns
        --------
        missing (list): list of file paths that are not in the g3tsmurf database
        """
        if session is None:
            session = self.Session()
        path = os.path.join(self.archive_path, str(timecode))

        q = session.query(Files).filter(Files.name.like(f"{path}%"))
        db_list = [f.name for f in q.all()]
        sys_list = walk_files(path)
        missing = []
        for f in sys_list:
            if f not in db_list:
                missing.append(f)

        return missing

    def find_missing_files_from_obs(self, timecode, session=None):
        """create a list of files in the g3tsmurf database that do not have an
        assigned level 2 observation ID
        
        Arguments
        ----------
        timecode (int): a level 2 timestreams timecode

        Returns
        --------
        missing (list): list of file paths that do not have level 2 observation IDs
        """
        if session is None:
            session = self.Session()
        path = os.path.join(self.archive_path, str(timecode))
        q = session.query(Files).filter(Files.name.like(f"{path}%"))
        db_list = q.all()
        return [f.name for f in db_list if f.obs_id is None]
        
    def lookup_file(self, filename, fail_ok=False):
        """Lookup a file's observations details in database. Meant to look
        and act like core.metadata.obsfiledb.lookup_file.
        """
        session = self.Session()
        file = session.query(Files).filter(Files.name == filename).one_or_none()

        if file is None and fail_ok:
            logger.debug(f"Did not find file {filename} in database")
            return None

        return {
            "obs_id": file.obs_id,
            "detsets": [file.detset],
            "sample_range": (file.sample_start, file.sample_stop),
        }

    def _stream_ids_in_range(self, start, end):
        """
        Returns a list of all stream-id's present in a given time range.
        Skips 'None' because those only contain G3PipelineInfo frames.

        Args
        -----
            start : timestamp or DateTime
                start time for data, assumed to be in UTC unless specified
            end :  timestamp or DateTime
                end time for data, assumed to be in UTC unless specified
        Returns
        --------
            stream_ids: List of stream ids.
        """
        session = self.Session()
        start = self._make_datetime(start)
        end = self._make_datetime(end)
        all_ids = (
            session.query(Files.stream_id)
            .filter(Files.start < end, Files.stop >= start)
            .all()
        )
        sids = []
        for (sid,) in all_ids:
            if sid not in sids and sid != "None":
                sids.append(sid)
        return sids

    def load_data(
        self,
        start,
        end,
        stream_id=None,
        channels=None,
        show_pb=True,
        load_biases=True,
        status=None,
    ):
        """
        Loads smurf G3 data for a given time range. For the specified time range
        this will return a chunk of data that includes that time range.

        This function returns an AxisManager with the following properties::

            * Axes:
                * samps : samples
                * dets : resonator channels reading out
                * bias_lines (optional) : bias lines

            * Fields:
               * timestamps : (samps,)
                    unix timestamps for loaded data
                * signal : (dets, samps)
                    Array of the squid phase in units of radians for each channel
                * primary : AxisManager (samps,)
                    "primary" data included in the packet headers
                    'AveragingResetBits', 'Counter0', 'Counter1', 'Counter2',
                    'FluxRampIncrement', 'FluxRampOffset', 'FrameCounter',
                    'TESRelaySetting', 'UnixTime'
                * biases (optional): (bias_lines, samps)
                    Bias values during the data
                * ch_info : AxisManager (dets,)
                    Information about channels, including SMuRF band, channel,
                    frequency.

        Args
        -----
            start : timestamp or DateTime
                start time for data, assumed to be in UTC unless specified
            end :  timestamp or DateTime
                end time for data, assumed to be in UTC unless specified
            stream_id : String
                stream_id to load, in case there are multiple
            channels : list or None
                If not None, it should be a list that can be sent to get_channel_mask.
            detset : string
                the name of the detector set (tuning file) to load
            show_pb : bool, optional:
                If True, will show progress bar.
            load_biases : bool, optional
                If True, will return biases.
            status : SmurfStatus, optional
                If note none, will use this Status on the data load

        Returns
        --------
            aman : AxisManager
                AxisManager for the data

        """
        session = self.Session()
        start = self._make_datetime(start)
        end = self._make_datetime(end)

        if stream_id is None:
            sids = self._stream_ids_in_range(start, end)
            if len(sids) > 1:
                raise ValueError(
                    "Multiple stream_ids exist in the given range! "
                    "Must choose one.\n"
                    f"stream_ids: {sids}"
                )

        q = (
            session.query(Files)
            .join(Frames)
            .filter(
                Frames.stop >= start, Frames.start < end, Frames.type_name == "Scan"
            )
        )
        if stream_id is not None:
            q = q.filter(Files.stream_id == stream_id)

        q = q.order_by(Files.start)
        flist = np.unique([x.name for x in q.all()])
        if stream_id is None:
            stream_id = q[0].stream_id

        if status is None:
            scan_start = session.query(Frames.time).filter(
                Frames.time >= start, Frames.type_name == "Scan"
            )
            scan_start = scan_start.order_by(Frames.time).first()

            try:
                status = self.load_status(scan_start[0], stream_id=stream_id)
            except:
                logger.info("Status load from database failed, using file load")
                status = None

        aman = load_file(
            flist,
            status=status,
            channels=channels,
            archive=self,
            show_pb=show_pb,
        )

        msk = np.all(
            [aman.timestamps >= start.timestamp(), aman.timestamps < end.timestamp()],
            axis=0,
        )
        idx = np.where(msk)[0]
        if len(idx) == 0:
            logger.warning("No samples returned in time range")
            aman.restrict("samps", (0, 0))
        else:
            aman.restrict("samps", (idx[0], idx[-1]))
        session.close()

        return aman

    def load_status(self, time, stream_id=None, show_pb=False):
        """
        Returns the status dict at specified unix timestamp.
        Loads all status frames between session start frame and specified time.

        Args:
            time (timestamp): Time at which you want the rogue status

        Returns:
            status (SmurfStatus instance): object indexing of rogue variables
            at specified time.
        """
        return SmurfStatus.from_time(time, self, stream_id=stream_id, show_pb=show_pb)

def dump_DetDb(archive, detdb_file):
    """
    Take a G3tSmurf archive and create a a DetDb of the type used with Context

    Args
    -----
        archive : G3tSmurf instance
        detdb_file : filename
    """
    my_db = core.metadata.DetDb(map_file=detdb_file)
    my_db.create_table("base", column_defs=[])
    column_defs = [
        "'band' int",
        "'channel' int",
        "'frequency' float",
        "'chan_assignment' int",
    ]
    my_db.create_table("smurf", column_defs=column_defs)

    ddb_list = my_db.dets()["name"]
    session = archive.Session()
    channels = session.query(Channels).all()
    msk = np.where([ch.name not in ddb_list for ch in channels])[0].astype(int)
    for ch in tqdm(np.array(channels)[msk]):
        my_db.get_id(name=ch.name)
        my_db.add_props(
            "smurf",
            ch.name,
            band=ch.band,
            channel=ch.channel,
            frequency=ch.frequency,
            chan_assignment=ch.chan_assignment.ctime,
        )
    session.close()
    return my_db

def make_DetDb_single_obs(obsfiledb, obs_id):
    # find relevant files to get status
    c = obsfiledb.conn.execute(
        "select name from files " "where obs_id=?" + "order by start", (obs_id,)
    )

    flist = [row[0] for row in c]

    # load status
    # for now, assume this is always the start of an obs so this won't break
    status = SmurfStatus.from_file(flist[0])

    # Pulling very specific pieces from get_channel_info. This isn't the best way to do this
    # but just want something that minimally works right now and I'll clean it up for PR

    # this should not be hardcoded but it is for now
    mask = None

    ch_list = np.arange(status.num_chans)
    ch_map = np.zeros(
        len(ch_list),
        dtype=[
            ("idx", int),
            ("rchannel", np.str_, 30),
            ("band", int),
            ("channel", int),
            ("freqs", float),
        ],
    )

    ch_map["idx"] = ch_list
    ch_map = _get_channel_mapping(status, ch_map)
    ruids = _get_detset_channel_names(status, ch_map, obsfiledb)

    # right now, require specific path to detdb, no helping userf
    detdb = core.metadata.DetDb()

    base_defs = [
        "'readout_id' str",
    ]
    column_defs = [
        "'band' int",
        "'channel' int",
        "'res_frequency' float",
    ]

    detdb.create_table("base", base_defs)
    detdb.create_table("smurf", column_defs)

    for i, ch in tqdm(enumerate(ch_map)):
        detdb.get_id(ruids[i])
        detdb.add_props("base", ruids[i], readout_id=ruids[i], commit=False)
        detdb.add_props(
            "smurf",
            ruids[i],
            band=ch["band"].item(),
            channel=ch["channel"].item(),
            res_frequency=ch["freqs"].item(),
            commit=False,
        )

    detdb.conn.commit()
    return detdb

def obs_detdb_context_hook(ctx, obs_id, *args, **kwargs):
    ddb = make_DetDb_single_obs(ctx.obsfiledb, obs_id)
    ctx.obs_detdb = ddb
    return ddb

core.Context.hook_sets["obs_detdb_load"] = {
    "before-use-detdb": obs_detdb_context_hook,
}

class SmurfStatus:
    """
    This is a class that attempts to extract essential information from the
    SMuRF status dictionary so it is more easily accessible. If the necessary
    information for an attribute is not present in the dictionary, the
    attribute will be set to None.

    Args
    -----
        status  : dict
            A SMuRF status dictionary

    Attributes
    ------------
        status : dict
            Full smurf status dictionary
        dump_frame : bool
            Whether the SmurfStatus was built off a dump frame
        num_chans: int
            Number of channels that are streaming
        mask : Optional[np.ndarray]
            Array with length ``num_chans`` that describes the mapping
            of readout channel to absolute smurf channel.
        mask_inv : np.ndarray
            Array with dimensions (NUM_BANDS, CHANS_PER_BAND) where
            ``mask_inv[band, chan]`` tells you the readout channel for a given
            band, channel combination.
        freq_map : Optional[np.ndarray]
            An array of size (NUM_BANDS, CHANS_PER_BAND) that has the mapping
            from (band, channel) to resonator frequency. If the mapping is not
            present in the status dict, the array will full of np.nan.
        filter_a : Optional[np.ndarray]
            The A parameter of the readout filter.
        filter_b : Optional[np.ndarray]
            The B parameter of the readout filter.
        filter_gain : Optional[float]
            The gain of the readout filter.
        filter_order : Optional[int]
            The order of the readout filter.
        filter_enabled : Optional[bool]
            True if the readout filter is enabled.
        downsample_factor : Optional[int]
            Downsampling factor
        downsample_enabled : Optional[bool]
            Whether downsampler is enabled
        flux_ramp_rate_hz : float
            Flux Ramp Rate calculated from the RampMaxCnt and the digitizer
            frequency.
        aman : AxisManager
            AxisManager containing the following fields:
                filter_a, filter_b, filter_gain, filter_order, filter_enabled,
                downsample_factor, flux_ramp_rate_hz
    """

    NUM_BANDS = 8
    CHANS_PER_BAND = 512

    def __init__(self, status):
        self.status = status
        self.start = self.status.get("start")
        self.stop = self.status.get("stop")
        self.stream_id = self.status.get("stream_id")
        self.session_id = self.status.get("session_id")

        self.aman = core.AxisManager()

        self.dump_frame = self.status.get("dump_frame")
        self.aman = self.aman.wrap("dump_frame", self.dump_frame)

        # Reads in useful status values as attributes
        mapper_root = "AMCc.SmurfProcessor.ChannelMapper"
        self.num_chans = self.status.get(f"{mapper_root}.NumChannels")

        # Tries to set values based on expected rogue tree
        self.mask = self.status.get(f"{mapper_root}.Mask")
        self.mask_inv = np.full((self.NUM_BANDS, self.CHANS_PER_BAND), -1)
        if self.mask is not None:
            self.mask = np.array(ast.literal_eval(self.mask))

            # Creates inverse mapping
            for i, chan in enumerate(self.mask):
                b = chan // self.CHANS_PER_BAND
                c = chan % self.CHANS_PER_BAND
                self.mask_inv[b, c] = i

        tune_root = "AMCc.FpgaTopLevel.AppTop.AppCore.SysgenCryo.tuneFilePath"
        self.tune = self.status.get(tune_root)
        if self.tune is not None and len(self.tune) > 0:
            self.tune = self.tune.split("/")[-1]

        pysmurf_root = "AMCc.SmurfProcessor.SOStream"
        self.action = self.status.get(f"{pysmurf_root}.pysmurf_action")
        if self.action == "":
            self.action = None
        self.action_timestamp = self.status.get(
            f"{pysmurf_root}.pysmurf_action_timestamp"
        )
        if self.action_timestamp == 0:
            self.action_timestamp = None

        filter_root = "AMCc.SmurfProcessor.Filter"
        self.filter_a = self.status.get(f"{filter_root}.A")
        if self.filter_a is not None:
            self.filter_a = np.array(ast.literal_eval(self.filter_a))
            self.aman = core.AxisManager(
                self.aman, core.IndexAxis("filter_pars", len(self.filter_a))
            )
            self.aman = self.aman.wrap("filter_a", self.filter_a, [(0, "filter_pars")])
        self.filter_b = self.status.get(f"{filter_root}.B")
        if self.filter_b is not None:
            self.filter_b = np.array(ast.literal_eval(self.filter_b))
            self.aman = self.aman.wrap("filter_b", self.filter_b, [(0, "filter_pars")])
        self.filter_gain = self.status.get(f"{filter_root}.Gain")
        self.aman = self.aman.wrap("filter_gain", self.filter_gain)
        self.filter_order = self.status.get(f"{filter_root}.Order")
        self.aman = self.aman.wrap("filter_order", self.filter_order)
        self.filter_enabled = not self.status.get("{filter_root}.Disable")
        self.aman = self.aman.wrap("filter_enabled", self.filter_enabled)

        ds_root = "AMCc.SmurfProcessor.Downsampler"
        self.downsample_factor = self.status.get(f"{ds_root}.Factor")
        self.aman = self.aman.wrap("downsample_factor", self.downsample_factor)
        self.downsample_enabled = not self.status.get(f"{ds_root}.Disable")

        ds_mode = self.status.get(f"{ds_root}.DownsamplerMode", "Internal")
        self.downsample_external = ds_mode.lower() == 'external'

        # Tries to make resonator frequency map
        self.freq_map = np.full((self.NUM_BANDS, self.CHANS_PER_BAND), np.nan)
        band_roots = [
            f"AMCc.FpgaTopLevel.AppTop.AppCore.SysgenCryo.Base[{band}]"
            for band in range(self.NUM_BANDS)
        ]
        for band in range(self.NUM_BANDS):
            band_root = band_roots[band]
            band_center = self.status.get(f"{band_root}.bandCenterMHz")
            subband_offset = self.status.get(f"{band_root}.toneFrequencyOffsetMHz")
            channel_offset = self.status.get(
                f"{band_root}.CryoChannels.centerFrequencyArray"
            )

            # Skip band if one of these fields is None
            if None in [band_center, subband_offset, channel_offset]:
                continue

            subband_offset = np.array(ast.literal_eval(subband_offset))
            channel_offset = np.array(ast.literal_eval(channel_offset))
            self.freq_map[band] = band_center + subband_offset + channel_offset

        # Calculates flux ramp reset rate (Pulled from psmurf's code)
        rtm_root = "AMCc.FpgaTopLevel.AppTop.AppCore.RtmCryoDet"
        ramp_max_cnt = self.status.get(f"{rtm_root}.RampMaxCnt")
        if ramp_max_cnt is None:
            self.flux_ramp_rate_hz = None
        else:
            digitizer_freq_mhz = float(
                self.status.get(f"{band_roots[0]}.digitizerFrequencyMHz", 614.4)
            )
            ramp_max_cnt_rate_hz = 1.0e6 * digitizer_freq_mhz / 2.0
            self.flux_ramp_rate_hz = ramp_max_cnt_rate_hz / (ramp_max_cnt + 1)
        self.aman = self.aman.wrap("flux_ramp_rate_hz", self.flux_ramp_rate_hz)
        self._make_tags()

    def _make_tags(self, delimiters=",|\\t| "):
        """Build list of tags from SMuRF status"""
        tags = self.status.get("AMCc.SmurfProcessor.SOStream.stream_tag")
        if tags is None:
            self.tags = []
            return
        self.tags = re.split(delimiters, tags)
        if len(self.tags) == 1 and self.tags[0] == "":
            self.tags = []

    @classmethod
    def from_file(cls, filename):
        """Generates a Smurf Status from a .g3 file.

        Args
        ----
            filename : str or list
        """
        if isinstance(filename, str):
            filenames = [filename]
        else:
            filenames = filename
        processor = cls._get_frame_processor()
        for file in filenames:
            reader = so3g.G3IndexedReader(file)
            while True:
                frames = reader.Process(None)
                if len(frames) == 0:
                    break
                if processor.process(frames[0]):
                    break
        return processor.get_status()

    @classmethod
    def _get_frame_processor(cls):
        """Returns a processor that can receive frames and ultimately return a
        SmurfStatus object.

        User should pass each frame into the return object's "process"
        method, then call get_status to return a SmurfStatus.  The
        "process" function returns True if the frame is a
        "dump_frame".

        """
        class _WiringFrameHandler(dict):
            def process(self, frame):
                status = self
                if str(frame.type) == "Wiring":
                    status["stream_id"] = frame.get("sostream_id")
                    status["session_id"] = frame.get("session_id")
                    if status.get("start") is None:
                        status["start"] = frame["time"].time / spt3g_core.G3Units.s
                        status["stop"] = frame["time"].time / spt3g_core.G3Units.s
                    else:
                        status["stop"] = frame["time"].time / spt3g_core.G3Units.s
                    status.update(yaml.safe_load(frame["status"]))
                    if frame["dump"]:
                        status["dump_frame"] = True
                        return True
                    status["dump_frame"] = False
                return False
            def get_status(self):
                return cls(self)
        return _WiringFrameHandler()

    @classmethod
    def from_time(cls, time, archive, stream_id=None, show_pb=False):
        """Generates a Smurf Status at specified unix timestamp.
        Loads all status frames between session start frame and specified time.

        Args
        -------
            time : (timestamp)
                Time at which you want the rogue status
                Assumed to be in UTC unless specified
            archive : (G3tSmurf instance)
                The G3tSmurf archive to use to find the status
            show_pb : (bool)
                Turn on or off loading progress bar
            stream_id : (string)
                stream_id to look for status

        Returns
        --------
            status : (SmurfStatus instance)
                object indexing of rogue variables at specified time.
        """
        time = archive._make_datetime(time)
        session = archive.Session()
        q = (
            session.query(Frames)
            .filter(Frames.type_name == "Observation", Frames.time <= time)
            .order_by(Frames.time.desc())
        )

        if stream_id is not None:
            q = q.join(Files).filter(Files.stream_id == stream_id)
        else:
            sids = archive._stream_ids_in_range(q[0].time, time)
            if len(sids) > 1:
                raise ValueError(
                    "Multiple stream_ids exist in the given range! "
                    "Must choose one to load SmurfStatus.\n"
                    f"stream_ids: {sids}"
                )

        if q.count() == 0:
            logger.error(f"No Frames found before time: {time}, stream_id: {stream_id}")

        start_frame = q.first()
        session_start = start_frame.time
        if stream_id is None:
            stream_id = start_frame.file.stream_id

        status_frames = (
            session.query(Frames)
            .join(Files)
            .filter(
                Files.stream_id == stream_id,
                Frames.type_name == "Wiring",
                Frames.time >= session_start,
                Frames.time <= time,
            )
            .order_by(Frames.time)
        )

        # Look for the last dump frame if avaliable
        dump_frame = (
            status_frames.filter(Frames.status_dump)
            .order_by(Frames.time.desc())
            .first()
        )

        if dump_frame is not None:
            status_frames = [dump_frame]
        else:
            logger.info("Status dump frame not found, reading all status frames")
            status_frames = status_frames.all()

        status = {
            "start": status_frames[0].time.timestamp(),
            "stop": status_frames[-1].time.timestamp(),
            "stream_id": stream_id,
            "session_id": status_frames[0].get("session_id"),
        }
        cur_file = None
        for frame_info in tqdm(status_frames, disable=(not show_pb)):
            file = frame_info.file.name
            if file != cur_file:
                reader = so3g.G3IndexedReader(file)
                cur_file = file
            reader.Seek(frame_info.offset)
            frame = reader.Process(None)[0]
            status.update(yaml.safe_load(frame["status"]))

        session.close()
        return cls(status)

    def readout_to_smurf(self, rchan):
        """
        Converts from a readout channel number to (band, channel).

        Args
        -----
            rchans : int or List[int]
                Readout channel to convert. If a list or array is passed,
                this will return an array of bands and array of smurf channels.

        Returns
        --------
            band, channel : (int, int) or (List[int], List[int])
                The band, channel combination that is has readout channel
                ``rchan``.
        """
        abs_smurf_chan = self.mask[rchan]
        return (
            abs_smurf_chan // self.CHANS_PER_BAND,
            abs_smurf_chan % self.CHANS_PER_BAND,
        )

    def smurf_to_readout(self, band, chan):
        """
        Converts from (band, channel) to a readout channel number.
        If the channel is not streaming, returns -1.

        Args:
            band : int, List[int]
                The band number, or list of band numbers corresopnding to
                channel input array.
            chan : int, List[int]
                Channel number or list of channel numbers.
        """
        return self.mask_inv[band, chan]

def get_channel_mask(
    ch_list, status, archive=None, obsfiledb=None, ignore_missing=True
):
    """Take a list of desired channels and parse them so the different
    data loading functions can load them.

    Args
    ------
    ch_list : list
        List of desired channels the type of each list element is used
        to determine what it is:

        * int : index of channel in file. Useful for batching.
        * (int, int) : band, channel
        * string : channel name (requires archive or obsfiledb)
        * float : frequency in the smurf status (or should we use channel assignment?)

    status : SmurfStatus instance
        Status to use to generate channel loading mask
    archive : G3tSmurf instance
        Archive used to search for channel names / frequencies
    obsfiledb : ObsFileDb instance
        ObsFileDb used to search for channel names if archive is None
    ignore_missing : bool
        If true, will not raise errors if a requested channel is not found

    Returns
    -------
    mask : bool array
        Mask for the channels in the SmurfStatus

    TODO: When loading from name, need to check tune file in use during file.
    """
    if status.mask is None:
        raise ValueError("Status Mask not set")

    session = None
    if archive is not None:
        session = archive.Session()

    msk = np.zeros((status.num_chans,), dtype="bool")
    for ch in ch_list:
        if np.isscalar(ch):
            if np.issubdtype(type(ch), np.integer):
                # this is an absolute readout INDEX (not band*512+ch)
                if not ignore_missing and ch >= status.num_chans:
                    raise ValueError(f"Requested Index {ch} > {status.num_chans}")
                if ch >= status.num_chans:
                    continue
                msk[ch] = True

            elif np.issubdtype(type(ch), np.floating):
                # this is a resonator frequency
                b, c = np.where(np.isclose(status.freq_map, ch, rtol=1e-7))
                if len(b) == 0:
                    if not ignore_missing:
                        raise ValueError(f"channel {ch} not found")
                    continue
                elif status.mask_inv[b, c][0] == -1:
                    if not ignore_missing:
                        raise ValueError(f"channel {ch} not streaming")
                    continue
                msk[status.mask_inv[b, c][0]] = True

            elif np.issubdtype(type(ch), np.str_):
                # this is a channel name
                if session is not None:
                    channel = (
                        session.query(Channels)
                        .filter(Channels.name == ch)
                        .one_or_none()
                    )
                    if channel is None:
                        if not ignore_missing:
                            raise ValueError(
                                f"channel {ch} not found in G3tSmurf Archive"
                            )
                        continue
                    b, c = channel.band, channel.channel
                elif obsfiledb is not None:
                    c = obsfiledb.conn.execute(
                        "select band,channel from channels where name=?", (ch,)
                    )
                    c = [(r[0], r[1]) for r in c]
                    if len(c) == 0:
                        if not ignore_missing:
                            raise ValueError(f"channel {ch} not found in obsfiledb")
                        continue
                    b, c = c[0]
                else:
                    raise ValueError(
                        "Need G3tSmurf Archive or Obsfiledb to pass channel names"
                    )

                idx = status.mask_inv[b, c]
                if idx == -1:
                    if not ignore_missing:
                        raise ValueError(f"channel {ch} not streaming")
                    continue
                msk[idx] = True

            else:
                raise TypeError(f"type {type(ch)} for channel {ch} not understood")
        else:
            if len(ch) == 2:
                # this is a band, channel pair
                idx = status.mask_inv[ch[0], ch[1]]
                if idx == -1:
                    if not ignore_missing:
                        raise ValueError(f"channel {ch} not streaming")
                    continue
                msk[idx] = True
            else:
                raise TypeError(f"type for channel {ch} not understood")
    if session is not None:
        session.close()
    return msk


def _get_tuneset_channel_names(status, ch_map, archive):
    """Update channel maps with name from Tuneset

    Returns
    ruids: list or None
        a list of readout_ids that align with ch_map. Returns None if
        readout_ids cannot be found.
    """
    session = archive.Session()

    # tune file in status
    if status.tune is not None and len(status.tune) > 0:
        tune_file = Tunes.get_name_from_status(status)
        tune = (
            session.query(Tunes)
            .filter(
                db.or_(Tunes.name == tune_file, Tunes.name == status.tune),
                Tunes.stream_id == status.stream_id,
            )
            .one_or_none()
        )
        if tune is None:
            logger.warning(f"Tune file {tune_file} not found in G3tSmurf archive")
            return None
        if tune.tuneset is None:
            logger.warning(f"Tune file {tune_file} has no TuneSet in G3tSmurf archive")
            return None
    else:
        logger.info("Tune information not in SmurfStatus, using most recent Tune")
        tune = session.query(Tunes).filter(
            Tunes.start <= dt.datetime.utcfromtimestamp(status.start),
            Tunes.stream_id == status.stream_id,
        )
        tune = tune.order_by(db.desc(Tunes.start)).first()
        if tune is None:
            logger.warning("Most recent Tune does not exist")
            return None
        if tune.tuneset is None:
            logger.warning(f"Tune file {tune.name} has no TuneSet in G3tSmurf archive")
            return None

    bands, channels, names = zip(
        *[(ch.band, ch.channel, ch.name) for ch in tune.tuneset.channels]
    )
    ruids = []

    for i in range(len(ch_map)):
        try:
            msk = np.all(
                [ch_map["band"][i] == bands, ch_map["channel"][i] == channels], axis=0
            )
            j = np.where(msk)[0][0]
            ruids.append(names[j])
        except:
            logger.info(f"Information retrival error for Detector {ch_map[i]}")
            ruids.append(
                "sch_NONE_{}_{:03d}".format(ch_map["band"][i], ch_map["channel"][i])
            )
            continue

    session.close()
    return ruids

def _get_detset_channel_names(status, ch_map, obsfiledb):
    """Update channel maps with name from obsfiledb"""
    # tune file in status
    if status.tune is not None and len(status.tune) > 0:
        c = obsfiledb.conn.execute(
            "select tuneset_id from tunes " "where name=?",
            (Tunes.get_name_from_status(status),),
        )
        temp = [r[0] for r in c]

        if len(temp) == 0:
            ## check for backward compatibility with old naming scheme. May 2023
            c = obsfiledb.conn.execute(
                "select tuneset_id from tunes " "where name=?", (status.tune,)
            )
            temp = [r[0] for r in c]
            if len(temp) == 0:
                logger.error(f"Tuneset id for {status.tune} not found")

        tuneset_id = temp[0]

    else:
        logger.info("Tune information not in SmurfStatus, using most recent Tune")
        c = obsfiledb.conn.execute(
            "select tuneset_id from tunes " "where start<=? " "order by start desc",
            (dt.datetime.utcfromtimestamp(status.start),),
        )
        tuneset_id = [r[0] for r in c][0]

    c = obsfiledb.conn.execute("select name from tunesets " "where id=?", (tuneset_id,))
    tuneset = [r[0] for r in c][0]

    c = obsfiledb.conn.execute("select det from detsets " "where name=?", (tuneset,))
    detsets = [r[0] for r in c]

    ruids = []

    if len(detsets) == 0:
        logger.warning(
            "Found no detsets related to this observation, is the database incomplete?"
        )

        for i in range(len(ch_map)):
            ruids.append(
                "sch_NONE_{}_{:03d}".format(ch_map["band"][i], ch_map["channel"][i])
            )
        return ruids

    sql = "select band,channel,name from channels where name in ({seq})".format(
        seq=",".join(["?"] * len(detsets))
    )
    c = obsfiledb.conn.execute(sql, detsets)
    bands, channels, names = zip(*[(r[0], r[1], r[2]) for r in c])

    for i in range(len(ch_map)):
        try:
            msk = np.all(
                [ch_map["band"][i] == bands, ch_map["channel"][i] == channels], axis=0
            )
            j = np.where(msk)[0][0]
            ruids.append(names[j])
        except:
            logger.info(f"Information retrival error for Detector {ch_map[i]}")
            ruids.append(
                "sch_NONE_{}_{:03d}".format(ch_map["band"][i], ch_map["channel"][i])
            )
            continue

    return ruids

def _get_channel_mapping(status, ch_map):
    """Generate baseline channel map from status object"""
    for i, ch in enumerate(ch_map["idx"]):
        try:
            sch = status.readout_to_smurf(ch)
            ch_map[i]["rchannel"] = "r{:04d}".format(ch)
            ch_map[i]["freqs"] = status.freq_map[sch[0], sch[1]]
            ch_map[i]["band"] = sch[0]
            ch_map[i]["channel"] = sch[1]
        except:
            ch_map[i]["rchannel"] = "r{:04d}".format(ch)
            ch_map[i]["freqs"] = -1
            ch_map[i]["band"] = -1
            ch_map[i]["channel"] = -1
    return ch_map

def get_channel_info(
    status,
    mask=None,
    archive=None,
    obsfiledb=None,
    det_axis="dets",
):
    """Create the Channel Info Section of a G3tSmurf AxisManager

    This function returns an AxisManager with the following properties::

        * Axes:
            * channels : resonator channels reading out

        * Fields:
            * band : Smurf Band
            * channel : Smurf Channel
            * frequency : resonator frequency
            * rchannel : readout channel
            * readout_id : readout unique ID

    Args
    -----
    status : SmurfStatus instance
    mask : bool or int array
        mask to select which channels to use (and possibly re-order them)
    archive : G3tSmurf instance (optionl)
        G3tSmurf instance for looking for tunes/tunesets
    obsfiledb : ObsfileDb instance (optional)
        ObsfileDb instance for det names / band / channel

    Returns
    --------
    ch_info : AxisManager

    """
    ch_list = np.arange(status.num_chans)
    if mask is not None:
        ch_list = ch_list[mask]

    ch_map = np.zeros(
        len(ch_list),
        dtype=[
            ("idx", int),
            ("rchannel", np.str_, 30),
            ("band", int),
            ("channel", int),
            ("freqs", float),
        ],
    )
    ch_map["idx"] = ch_list

    ch_map = _get_channel_mapping(status, ch_map)

    if archive is not None:
        ruids = _get_tuneset_channel_names(status, ch_map, archive)
    elif obsfiledb is not None:
        ruids = _get_detset_channel_names(status, ch_map, obsfiledb)
    else:
        ruids = None

    if ruids is None:
        labels = [
            "sbch_{}_{:03d}".format(ch_map["band"][i], ch_map["channel"][i])
            for i in range(len(ch_list))
        ]
        ch_info = core.AxisManager(
            core.LabelAxis(det_axis, labels),
        )
    else:
        ch_info = core.AxisManager(
            core.LabelAxis(det_axis, ruids),
        )

    ch_info.wrap("band", ch_map["band"], ([(0, det_axis)]))
    ch_info.wrap("channel", ch_map["channel"], ([(0, det_axis)]))
    ch_info.wrap("frequency", ch_map["freqs"], ([(0, det_axis)]))
    ch_info.wrap("rchannel", ch_map["rchannel"], ([(0, det_axis)]))
    if ruids is not None:
        ch_info.wrap("readout_id", np.array(ruids), ([(0, det_axis)]))

    return ch_info

def _get_sample_info(filenames):
    """Scan through a list of files and count samples. Starts counting
    from the first file in the list. Used in load_file for sample restiction
    if no database connection is available.

    Args
    -----
    filenames : list
        list of filenames

    Returns
    --------
    out : list
        a list of dictionaries formatted for load_file.
        in the pattern of [ {filename: filename,
                            sample_range: (sample_start, sample_stop)}]
    """
    out = []
    start = 0
    for file in filenames:
        samps = 0
        reader = so3g.G3IndexedReader(file)
        while True:
            frames = reader.Process(None)
            if len(frames) == 0:
                break
            frame = frames[0]
            if str(frame.type) != "Scan":
                continue
            data = frame.get("data")
            sostream_version = frame.get("sostream_version", 0)
            if sostream_version >= 2:
                samps += len(data.times)
            else:
                samps += len(data)
        out.append({"filename": file, "sample_range": (start, start + samps)})
        start += samps
    return out

def split_ts_bits(c):
    """Split up 64 bit to 2x32 bit"""
    NUM_BITS_PER_INT = 32
    MAXINT = (1 << NUM_BITS_PER_INT) - 1
    a = (c >> NUM_BITS_PER_INT) & MAXINT
    b = c & MAXINT
    return a, b

def _get_timestamps(streams, load_type=None, linearize_timestamps=True):
    """Calculate the timestamp field for loaded data

    Args
    -----
        streams : dictionary
            result from unpacking the desired data frames
        load_type : None or int
            if None, uses highest precision version possible. integer values
            will use the TimingParadigm class for indexing
        linearize_timestamps : bool
          if true and using unix timing, linearize the timing based on the
          frame counter
    """
    if load_type is None:
        # determine the desired loading type. Expand as logic as
        # data fields develop
        if "primary" in streams:
            if (
                np.abs(
                    np.diff(io_load.hstack_into(None, streams["primary"]["Counter0"]))
                ).mean()
                != 0
            ):
                load_type = TimingParadigm.TimingSystem
            elif "UnixTime" in streams["primary"]:
                load_type = TimingParadigm.SmurfUnixTime
            else:
                load_type = TimingParadigm.G3Timestream
        else:
            load_type = TimingParadigm.G3Timestream

    if load_type == TimingParadigm.TimingSystem:
        s, ns = split_ts_bits(io_load.hstack_into(None, streams["primary"]["Counter2"]))
        # Add 20 years in seconds (accounting for leap years) to handle
        # offset between EPICS time referenced to 1990 relative to UNIX time.
        counter2 = s + ns * 1e-9 + 5 * (4 * 365 + 1) * 24 * 60 * 60
        counter0 = io_load.hstack_into(None, streams["primary"]["Counter0"])
        timestamps = np.round(counter2 - (counter0 / 480000)) + counter0 / 480000
        return timestamps
    if load_type == TimingParadigm.SmurfUnixTime:
        timestamps = io_load.hstack_into(None, streams["primary"]["UnixTime"]) / 1e9
        if linearize_timestamps:
            frames = io_load.hstack_into(None, streams["primary"]["FrameCounter"])
            fsamp, offset = np.polyfit(frames, timestamps, 1)
            timestamps = offset + fsamp * frames
        return timestamps
    if load_type == TimingParadigm.G3Timestream:
        return io_load.hstack_into(None, streams["time"])
    logger.error("Timing System could not be determined")

def load_file(
    filename,
    channels=None,
    samples=None,
    ignore_missing=True,
    no_signal=False,
    load_biases=True,
    load_primary=True,
    status=None,
    archive=None,
    obsfiledb=None,
    show_pb=True,
    det_axis="dets",
    linearize_timestamps=True,
    merge_det_info=True,
):
    """Load data from file where there may or may not be a connected archive.

    Args
    ----
      filename : str or list
          A filename or list of filenames (to be loaded in order).
          Note that SmurfStatus is only loaded from the first file
      channels: list or None
          If not None, it should be a list that can be sent to get_channel_mask.
      samples : tuple or None
          If not None, it should be a tuple of (sample_start, sample_stop) where the
          sample counts are relative to the entire g3 session, not just the files
          being loaded from the list.
      ignore_missing : bool
          If true, will not raise errors if a requested channel is not found
      no_signal : bool
          If true, will not load the detector signal from files
      load_biases : bool
          If true, will load the bias lines for each detector
      load_primary : bool
          If true, loads the primary data fields, old .g3 files may not have
          these fields.
      archive : a G3tSmurf instance (optional)
      obsfiledb : a ObsFileDb instance (optional, used when loading from context)
      status : a SmurfStatus Instance if we don't want to use the one from the
          first file
      det_axis : name of the axis used for channels / detectors
      linearize_timestamps : bool
          sent to _get_timestamps. if true and using unix timing, linearize the timing
          based on the frame counter
      merge_det_info : bool
          if true, emulate det_info from file info

    Returns
    ---------
      aman : AxisManager
        AxisManager with the data with axes for `channels` and `samps`. It will
        always have fields `timestamps`, `signal`, `flags`(FlagManager),
        `ch_info` (AxisManager with `bands`, `channels`, `frequency`, etc).
    """
    logger.debug(f"Axis Manager will have {det_axis} and samps axes")

    if isinstance(filename, str):
        filenames = [filename]
    else:
        filenames = filename

    if len(filenames) == 0:
        logger.error("No files provided to load")

    if status is not None and status.num_chans is None:
        logger.warning("Status information is missing 'num_chans.' Will try to fix.")
        status = None

    if status is None:
        try:
            logger.debug(f"Loading status from {filenames[0]}")
            status = SmurfStatus.from_file(filenames[0])
        except:
            logger.warning(f"Failed to load status from {filenames[0]}.")

    if status is None or status.num_chans is None:
        try:
            logger.warning(
                f"Complete status not available in {filenames[0]}\n"
                "Trying to load status frame from the file at the start "
                "of the corresponding observation."
            )

            file_id = filenames[0].split("/")[-1][10:]
            status_fp = filenames[0].replace(file_id, "_000.g3")
            status = SmurfStatus.from_file(status_fp)
        except Exception as e:
            logger.error(
                f"Error when trying to load status from {status_fp}, maybe the file doesn't exist?"
                "Please load the status manually."
            )
            raise e

    if channels is not None:
        if len(channels) == 0:
            logger.error(
                "Requested empty list of channels. Use channels=None to "
                "load all channels."
            )
        ch_mask = get_channel_mask(
            channels,
            status,
            archive=archive,
            obsfiledb=obsfiledb,
            ignore_missing=ignore_missing,
        )
        is_many_channels = ch_mask.sum() >= len(ch_mask) * 0.5
    else:
        ch_mask = None
        is_many_channels = True

    ch_info = get_channel_info(
        status,
        ch_mask,
        archive=archive,
        obsfiledb=obsfiledb,
        det_axis=det_axis,
    )

    # flist will take the form [(file, sample_start, sample_stop)...] and will be
    # passed to io_load.unpack_frames
    flist = []
    if samples is None:
        sample_start, sample_stop = 0, None
        flist = [(f, 0, None) for f in filenames]
    else:
        sample_start, sample_stop = samples

        if archive is None and obsfiledb is None:
            outs = _get_sample_info(filenames)
        else:
            X = [archive if archive is not None else obsfiledb][0]
            outs = [X.lookup_file(file) for file in filenames]
        stop = sample_stop
        for filename, out in zip(filenames, outs):
            file_start, file_stop = out["sample_range"]
            if file_stop <= sample_start:
                continue
            if stop is not None:
                if file_start >= sample_stop:
                    continue
                stop = sample_stop - file_start

            start = max(0, sample_start - file_start)
            flist.append((filename, start, stop))

    if no_signal:
        subreq = [io_load.FieldGroup("data", [], timestamp_field="time")]
    else:
        subreq = [
            io_load.FieldGroup(
                "data",
                ch_info.rchannel,
                timestamp_field="time",
                refs_ok=is_many_channels,
            )
        ]
    if load_primary:
        subreq.extend(
            [io_load.FieldGroup("primary", [io_load.Field("*", wildcard=True)])]
        )
    if load_biases:
        subreq.extend(
            [
                io_load.FieldGroup("tes_biases", [io_load.Field("*", wildcard=True)]),
            ]
        )

    request = io_load.FieldGroup("root", subreq)
    streams = None
    try:
        for filename, start, stop in tqdm(
            flist, total=len(flist), disable=(not show_pb)
        ):
            streams = io_load.unpack_frames(
                filename, request, streams=streams, samples=(start, stop)
            )
    except KeyError:
        logger.error(
            "Frames do not contain expected fields. Did Channel Mask change during the file?"
        )
        raise

    count = sum(map(len, streams["time"]))

    # Build AxisManager
    aman = core.AxisManager(
        ch_info[det_axis],
        core.OffsetAxis("samps", count, sample_start),
    )
    aman.wrap(
        "timestamps",
        _get_timestamps(streams, linearize_timestamps=linearize_timestamps),
        ([(0, "samps")]),
    )
    aman.wrap("status", status.aman)
    if merge_det_info:
        det_info = core.AxisManager(ch_info[det_axis])
        smurf = core.AxisManager(ch_info[det_axis])
        det_info.wrap("readout_id", ch_info[det_axis].vals, [(0, det_axis)])
        smurf.wrap("band", ch_info.band, [(0, det_axis)])
        smurf.wrap("channel", ch_info.channel, [(0, det_axis)])
        smurf.wrap("res_frequency", ch_info.frequency, [(0, det_axis)])
        det_info.wrap("smurf", smurf)
        aman.wrap("det_info", det_info)

    # If readout filter in enabled build iir_params AxisManager
    if status.filter_enabled:
        iir_params = core.AxisManager()
        iir_params.wrap("a", status.filter_a)
        iir_params.wrap("b", status.filter_b)
        iir_params.wrap("fscale", 1 / status.flux_ramp_rate_hz)
        aman.wrap("iir_params", iir_params)

    if not no_signal:
        aman.wrap(
            "signal",
            np.zeros((aman[det_axis].count, aman["samps"].count), "float32"),
            [(0, det_axis), (1, "samps")],
        )
        for idx in range(aman[det_axis].count):
            io_load.hstack_into(
                aman.signal[idx], streams["data"][ch_info.rchannel[idx]]
            )

        # Conversion from DAC counts to squid phase
        rad_per_count = np.pi / 2**15
        aman.signal *= rad_per_count

    temp = core.AxisManager(aman.samps.copy())
    if load_primary:
        for k in streams["primary"].keys():
            temp.wrap(
                k, io_load.hstack_into(None, streams["primary"][k]), ([(0, "samps")])
            )
        aman.wrap("primary", temp)

    if load_biases:
        bias_labels = [
            "b{:02d}".format(b_num)
            for b_num in range(len(streams["tes_biases"].keys()))
        ]
        bias_axis = core.LabelAxis("bias_lines", np.array(bias_labels))
        aman.wrap(
            "biases",
            np.zeros((bias_axis.count, aman.samps.count)),
            [(0, bias_axis), (1, "samps")],
        )
        for k in streams["tes_biases"].keys():
            i = int(k[4:])
            io_load.hstack_into(aman.biases[i], streams["tes_biases"][k])
    aman.wrap("flags", core.FlagManager.for_tod(aman, det_axis, "samps"))

    return aman

def load_g3tsmurf_obs(db, obs_id, dets=None, samples=None, no_signal=None, **kwargs):
    """Obsloader function for g3tsmurf data archives.

    See API template, `sotodlib.core.context.obsloader_template`, for
    details.

    """
    if any([v is not None for v in kwargs.values()]):
        raise RuntimeError(
            f"This loader function does not understand kwargs: f{kwargs}"
        )
    c = db.conn.execute(
        "select name from files " "where obs_id=?" + "order by start", (obs_id,)
    )
    flist = [row[0] for row in c]
    if no_signal is None:
        no_signal = False
    return load_file(
        flist,
        dets,
        samples=samples,
        obsfiledb=db,
        no_signal=no_signal,
        merge_det_info=False,
    )


core.OBSLOADER_REGISTRY["g3tsmurf"] = load_g3tsmurf_obs



#!/usr/bin/env python3
# Copyright (c) 2019-2021 Simons Observatory.
# Full license can be found in the top level "LICENSE" file.

"""
This script runs a basic mapmaking test.

DEPRECATED:  This script is left here for reference, but was only used for debugging.

"""

import os
import sys
import traceback
import argparse

import numpy as np

from astropy import units as u

# Import sotodlib.toast first before toast, since that sets default object names
# to use in toast.
import sotodlib.toast as sotoast
from sotodlib import mapmaking as mm

import toast
import toast.ops

from toast.mpi import MPI

from toast.schedule_sim_ground import run_scheduler

from .. import ops as so_ops

# Make sure pixell uses a reliable FFT engine
import pixell.fft
pixell.fft.engine = "fftw"


def parse_args(mlmapmaker, comm):
    """Parse command line arguments"""
    # Argument parsing
    parser = argparse.ArgumentParser(description="SO mapmaker test")

    parser.add_argument(
        "--hardware",
        required=True,
        default=None,
        help="Input hardware file, trimmed to desired detectors.",
    )
    parser.add_argument(
        "--schedule", required=True, default=None, help="Input schedule file."
    )
    parser.add_argument(
        "--sample_rate",
        required=False,
        default=100.0,
        type=float,
        help="Sampling rate.",
    )
    parser.add_argument(
        "--sky_file",
        required=True,
        default=None,
        help="Input NSIDE=4096 TQU file to scan from.",
    )
    parser.add_argument(
        "--out_dir",
        required=False,
        type=str,
        default="out_mapmaker_test",
        help="The output directory",
    )

    config, args, jobargs = toast.parse_config(parser, operators=[mlmapmaker])

    # Log the config that was actually used at runtime.
    outlog = os.path.join(args.out_dir, "config_log.toml")
    toast.config.dump_toml(outlog, config, comm=comm)

    return config, args, jobargs


def load_schedule(comm, path):
    """Create and load a schedule file."""
    schedule = toast.schedule.GroundSchedule()
    if comm is None or comm.rank == 0:
        if not os.path.isfile(path):
            raise RuntimeError(f"Schedule file {path} does not exist")
        schedule.read(path)
    if comm is not None:
        schedule = comm.bcast(schedule, root=0)
    return schedule


def load_instrument(comm, args, schedule):
    focalplane = sotoast.SOFocalplane(
        hwfile=args.hardware,
        telescope="LAT",
        sample_rate=args.sample_rate * u.Hz,
        comm=comm,
    )

    site = toast.instrument.GroundSite(
        schedule.site_name,
        schedule.site_lat,
        schedule.site_lon,
        schedule.site_alt,
        weather=None,
    )

    telescope = toast.instrument.Telescope(
        focalplane.telescope, focalplane=focalplane, site=site
    )
    return telescope


def main():
    env = toast.utils.Environment.get()
    log = toast.utils.Logger.get()
    gt = toast.timing.GlobalTimers.get()
    gt.start("mapmaker_test main")

    mlmapmaker = so_ops.MLMapmaker(comps="TQU")

    # Get optional MPI parameters
    comm, procs, rank = toast.get_world()

    # Get commandline options
    config, args, jobargs = parse_args(mlmapmaker, comm)
    job = toast.create_from_config(config)
    mlmapmaker = job.operators.MLMapmaker

    # Make the output directory
    if rank == 0:
        os.makedirs(args.out_dir, exist_ok=True)

    # Load (and optionally create) our schedule file in the output directory
    schedule = load_schedule(comm, args.schedule)

    # Load our instrument model
    telescope = load_instrument(comm, args, schedule)

    # Create the toast communicator.  The ML mapmaker currently requires
    # a single process to have all data in an observation.  So we create
    # groups of one process.
    toast_comm = toast.Comm(world=comm, groupsize=1)

    # Timer for reporting the progress
    timer = toast.timing.Timer()
    timer.start()

    # Shortcut for the world communicator
    wcomm = toast_comm.comm_world

    # Note on toast operators:  Operators are configured by "traits" from
    # the python traitlets package.  These can be set at construction time
    # or afterwards by just changing the value.  We use both techniques
    # below but either could be used exclusively.

    # Note on "names".  We use default names for most objects within an
    # observation (e.g. "signal", "elevation", "boresight_radec", etc).
    # These defaults can be changed globally- see sotodlib.toast.__init__.py .

    # ======================================================
    # Simulate Data
    # ======================================================

    # Create the (initially empty) data
    data = toast.Data(comm=toast_comm)

    # Simulate the telescope pointing
    sim_ground = toast.ops.SimGround(weather="atacama")
    sim_ground.telescope = telescope
    sim_ground.schedule = schedule
    sim_ground.apply(data)
    log.info_rank("Simulated telescope pointing in", comm=wcomm, timer=timer)

    # Apply LAT co-rotation
    corotate_lat = so_ops.CoRotator()
    corotate_lat.apply(data)
    log.info_rank("Applied LAT co-rotation in", comm=wcomm, timer=timer)

    # Construct a "perfect" noise model just from the focalplane parameters
    default_noise = toast.ops.DefaultNoiseModel()
    default_noise.apply(data)
    log.info_rank("Created default noise model in", comm=wcomm, timer=timer)

    # Configure detector pointing in both Az/El and RA/DEC.  We are not expanding
    # pointing- these classes are passed to other operators.

    det_pointing_azel = toast.ops.PointingDetectorSimple(quats="quats_azel")
    det_pointing_azel.boresight = sim_ground.boresight_azel

    det_pointing_radec = toast.ops.PointingDetectorSimple(quats="quats_radec")
    det_pointing_radec.boresight = sim_ground.boresight_radec

    # Create the Elevation modulated noise model
    elevation_noise = toast.ops.ElevationNoise(out_model="el_noise_model")
    elevation_noise.detector_pointing = det_pointing_azel
    elevation_noise.apply(data)
    log.info_rank("Created elevation noise model in", comm=wcomm, timer=timer)

    # Set up the pointing matrix in RA / DEC, and pointing weights in Az / El
    # in case we need that for the atmosphere sim below.
    pixels_radec = toast.ops.PixelsHealpix(
        detector_pointing=det_pointing_radec,
        nside=4096,
    )
    weights_radec = toast.ops.StokesWeights(weights="weights_radec", mode="IQU")
    weights_radec.detector_pointing = det_pointing_radec
    weights_azel = toast.ops.StokesWeights(weights="weights_azel", mode="IQU")
    weights_azel.detector_pointing = det_pointing_azel

    # Scan input map.  This will create the pixel distribution as well, since
    # it does not yet exist.
    scan_map = toast.ops.ScanHealpixMap(file=args.sky_file)
    scan_map.enabled = True
    scan_map.pixel_pointing = pixels_radec
    scan_map.stokes_weights = weights_radec
    scan_map.apply(data)
    log.info_rank("Simulated sky signal in", comm=wcomm, timer=timer)

    # Simulate atmosphere.  For this test script, we are just simulating one component
    # and excluding the "coarse" large scale component.
    sim_atmosphere = toast.ops.SimAtmosphere(
        lmin_center=0.001 * u.meter,
        lmin_sigma=0.0 * u.meter,
        lmax_center=1.0 * u.meter,
        lmax_sigma=0.0 * u.meter,
        gain=1.0e-4,
        zatm=40000 * u.meter,
        zmax=200 * u.meter,
        xstep=5 * u.meter,
        ystep=5 * u.meter,
        zstep=5 * u.meter,
        nelem_sim_max=10000,
        wind_dist=3000 * u.meter,
        z0_center=2000 * u.meter,
        z0_sigma=0 * u.meter,
        cache_dir="atm_cache",
    )
    sim_atmosphere.detector_pointing = det_pointing_azel
    # Here is where we could enable a small polarization fraction, in which
    # case we need to specify the Stokes weights in Az/El.
    # sim_atmosphere.polarization_fraction = 0.01
    # sim_atmosphere.detector_weights = weights_azel
    sim_atmosphere.enabled = False  # Toggle to False to disable
    sim_atmosphere.serial = False
    sim_atmosphere.apply(data)
    log.info_rank("Simulated and observed atmosphere in", comm=wcomm, timer=timer)

    # Simulate detector noise
    sim_noise = toast.ops.SimNoise()
    sim_noise.noise_model = elevation_noise.out_model
    sim_noise.serial = False
    sim_noise.apply(data)
    log.info_rank("Simulated detector noise in", comm=wcomm, timer=timer)

    # ======================================================
    # Reduce Data
    # ======================================================

    # Various geometric factors

    h_n = so_ops.Hn(pixel_pointing=pixels_radec, output_dir=args.out_dir)
    h_n.enabled = False  # Toggle to False to disable
    h_n.apply(data)
    log.info_rank("Calculated h_n in", comm=wcomm, timer=timer)

    cadence_map = toast.ops.CadenceMap(
        pixel_pointing=pixels_radec, output_dir=args.out_dir
    )
    cadence_map.enabled = False  # Toggle to False to disable
    cadence_map.apply(data)
    log.info_rank("Calculated cadence map in", comm=wcomm, timer=timer)

    crosslinking = toast.ops.CrossLinking(
        pixel_pointing=pixels_radec, output_dir=args.out_dir
    )
    crosslinking.enabled = False  # Toggle to False to disable
    crosslinking.apply(data)
    log.info_rank("Calculated crosslinking in", comm=wcomm, timer=timer)

    # Ground (scan synchronous signal) filter.  Remove modes that are poorly
    # constrained by the scanning.  Disable this if the MLMapmaker does
    # something similar internally.

    ground_filter = toast.ops.GroundFilter()
    ground_filter.enabled = False  # Toggle to False to disable
    ground_filter.apply(data)
    log.info_rank("Finished ground-filtering in", comm=wcomm, timer=timer)

    # Run ML mapmaker
    mlmapmaker.enabled = True  # Toggle to False to disable
    mlmapmaker.out_dir = args.out_dir
    mlmapmaker.apply(data)
    log.info_rank("Finished ML map-making in", comm=wcomm, timer=timer)

    # Filter and bin.  Apply a 2D polynomial filter and then make a binned map
    # We could also iteratively solve for other templates here, but skipping
    # that for now.

    polyfilter2D = toast.ops.PolyFilter2D()
    polyfilter2D.enabled = False  # Toggle to False to disable
    polyfilter2D.apply(data)
    log.info_rank("Finished 2D-poly-filtering in", comm=wcomm, timer=timer)

    binner = toast.ops.BinMap()
    binner.noise_model = elevation_noise.out_model
    binner.pixel_pointing = pixels_radec
    binner.stokes_weights = weights_radec

    mapmaker = toast.ops.MapMaker(name="mapmaker")
    mapmaker.weather = "vacuum"
    mapmaker.write_hdf5 = True
    mapmaker.binning = binner
    # No templates for now (will just do the binning)
    mapmaker.template_matrix = toast.ops.TemplateMatrix(templates=[])
    mapmaker.output_dir = args.out_dir
    mapmaker.enabled = False  # Toggle to False to disable
    mapmaker.apply(data)
    log.info_rank("Finished Toast map-making in", comm=wcomm, timer=timer)

    # Collect optional timing information
    alltimers = toast.timing.gather_timers(comm=toast_comm.comm_world)
    if toast_comm.world_rank == 0:
        out = os.path.join(args.out_dir, "timing")
        toast.timing.dump(alltimers, out)


def cli():
    world, procs, rank = toast.mpi.get_world()
    with toast.mpi.exception_guard(comm=world):
        main()


if __name__ == "__main__":
    cli()



"""
This submodule contains classes to perform filter+bin mapmaking for 
demodulated data. The DemodMapmaker class creates a mapmaker object.
DemodSignalMap creates a signal you want to solve for, over which 
you accumulate observations into the div and rhs maps. For examples
how to use look at docstring of DemodMapmaker.
"""
__all__ = ['DemodMapmaker','DemodSignal','DemodSignalMap','make_demod_map','setup_demod_map']
import numpy as np
from pixell import enmap, utils as putils, tilemap, bunch, mpi

from .. import coords
from . import utils as smutils
from .noise_model import NmatWhite

hp = smutils.import_optional('healpy')
h5py = smutils.import_optional('h5py')

class DemodMapmaker:
    def __init__(self, signals=[], noise_model=None, dtype=np.float32, verbose=False, comps='TQU', singlestream=False):
        """
        Initialize a FilterBin Mapmaker for demodulated data 
        
        Arguments
        ---------
        signals : list 
            List of Signal-objects representing the models that will be 
            solved jointly for. Currently this would be a DemodSignal
        noise_model : sotodlib.mapmaking.Nmat or None
            A noise model constructor which will be used to initialize the
            noise model for each observation. Can be overriden in add_obs.
            Noises other than NmatWhite not implemented. If None, a white 
            noise model is used.
        dtype : numpy.dtype, optional
            The data type to use for the time-ordered data. Only tested
            with float32
        verbose : Bool, optional
            Whether to print progress messages. Not implemented
        comps : str, optional
            String with the components to solve for. Not implemented for 
            anything other than TQU
        singlestream : Bool, optional
            If True, do not perform demodulated filter+bin mapmaking but 
            rather regular filter+bin mapmaking, i.e. map from obs.signal
            rather than from obs.dsT, obs.demodQ, obs.demodU
        
        Example usage :: 
            signal_map = mapmaking.DemodSignalMap(shape, wcs, comm)
            signals    = [signal_map]
            mapmaker   = mapmaking.DemodMapmaker(signals, 
                         noise_model=noise_model)
        
        """
        
        if noise_model is None:
            noise_model = NmatWhite()
        self.signals      = signals
        self.dtype        = dtype
        self.verbose      = verbose
        self.noise_model  = noise_model
        self.data         = []
        self.dof          = smutils.MultiZipper()
        self.ready        = False
        self.ncomp        = len(comps)
        self.singlestream = singlestream

    def add_obs(self, id, obs, noise_model=None, split_labels=None,
                use_psd=True, wn_label='preprocess.noiseQ_mapmaking.white_noise',
                apply_wobble=True):
        """
        This function will accumulate an obs into the DemodMapmaker object, i.e. will add to 
        a RHS and div map.
        
        Arguments
        ---------
        id : str
            String that identifies an observation to accumulate. Typically will be something 
            like obs_id:wafer_slot:band
        obs : AxisManager
            An observation (axismanager) object to be accumulated.
        noise_model : sotodlib.mapmaking.Nmat or None, optional
            If a noise model is passed, the model in the Demodmapmaker object will be overriden,
            if None, the default model defined in the Demodmapmaker object is used.
        split_labels: list or None, optional
            A list of strings with the splits requested. If None then no splits were asked for,
            i.e. we will produce one map
        use_psd : bool, optional
            By default True. Use the white noise measured from PSD as mapmaking weights,
            which must be provided in the preprocessing. This is done as opposed to
            build the ivar locally from the std of the TOD.
        wn_label : str, optional
            Path where to find the white noise per det estimated by the preprocessing.

        """
        ctime  = obs.timestamps
        srate  = (len(ctime)-1)/(ctime[-1]-ctime[0])
        if not(self.singlestream):
            # now we have 3 signals, dsT / demodQ / demodU. We pack them into an array with shape (3,...)
            tod    = np.array([obs.dsT.astype(self.dtype, copy=False), obs.demodQ.astype(self.dtype, copy=False), obs.demodU.astype(self.dtype, copy=False)])
        else:
            tod = obs.signal.astype(self.dtype, copy=False)
        # Allow the user to override the noise model on a per-obs level
        if noise_model is None: noise_model = self.noise_model
        # Build the noise model from the obs unless a fully
        # initialized noise model was passed
        if use_psd:
            try:
                noise_model.ivar = 1.0 / obs[wn_label]**2
            except:
                msg = f"use_psd is True but '{wn_label}' does not contain the white noise estimated on the preprocessing"
                raise RuntimeError(msg)
        if noise_model.ready:
            nmat = noise_model
        else:
            try:
                if not(self.singlestream):
                    nmat = noise_model.build(tod[1], srate=srate) # Here we are building the model from demodQ
                else:
                    nmat = noise_model.build(tod, srate=srate)
            except Exception as e:
                msg = f"FAILED to build a noise model for observation='{id}' : '{e}'"
                raise RuntimeError(msg)
        # Add the observation to each of our signals
        for signal in self.signals:
            signal.add_obs(id, obs, nmat, tod, split_labels=split_labels,apply_wobble=apply_wobble)
        # Save what we need about this observation
        self.data.append(bunch.Bunch(id=id, ndet=obs.dets.count, nsamp=len(ctime), dets=obs.dets.vals, nmat=nmat))

class DemodSignal:
    def __init__(self, name, ofmt, output, ext):
        """
        This class represents a thing we want to solve for, e.g. the sky, ground, cut samples, etc.
        
        Arguments
        ---------
        name : str
            The name of this signal, e.g. "sky", "cut", etc.
        ofmt : str
            The format used when constructing output file prefix
        output : Bool
            Whether this signal should be part of the output or not.
        ext : str
            The extension used for the files.

        """
        self.name   = name
        self.ofmt   = ofmt
        self.output = output
        self.ext    = ext
        self.dof    = None
        self.ready  = False
    def add_obs(self, id, obs, nmat, Nd): pass
    def prepare(self): self.ready = True
    def to_work  (self, x): return x.copy()
    def from_work(self, x): return x
    def write   (self, prefix, tag, x): pass

class DemodSignalMap(DemodSignal):
    def __init__(self, shape=None, wcs=None, comm=None, comps="TQU", name="sky", ofmt="{name}", output=True,
                 ext="fits", dtype_map=np.float64, dtype_tod=np.float32, sys=None, recenter=None, tile_shape=(500,500), tiled=False, Nsplits=1, singlestream=False,
                 nside=None, nside_tile=None):
        """
        Parent constructor; use for_rectpix or for_healpix instead. Args described there.
        """
        DemodSignal.__init__(self, name, ofmt, output, ext)
        self.comm  = comm
        self.comps = comps
        self.sys   = sys
        self.recenter = recenter
        self.dtype_map = dtype_map
        self.dtype_tod = dtype_tod
        self.tile_shape = tile_shape        
        self.tiled = tiled
        self.data  = {}
        self.Nsplits = Nsplits
        self.singlestream = singlestream
        self.wrapper = lambda x : x
        self.wcs = wcs
        ncomp      = len(comps)

        self.pix_scheme = "rectpix" if (wcs is not None) else "healpix"
        if self.pix_scheme == "healpix":
            self.tiled = (nside_tile is not None)
            self.hp_geom = coords.healpix_utils.get_geometry(nside, nside_tile, ordering='NEST')
            npix = 12 * nside**2
            self.rhs = np.zeros((Nsplits, ncomp, npix), dtype=dtype_map)
            self.div = np.zeros((Nsplits, ncomp, ncomp, npix), dtype=dtype_map)
            self.hits = np.zeros((Nsplits, npix), dtype=dtype_map)

            if self.tiled:
                self.wrapper = coords.healpix_utils.tiled_to_full
        else:
            if shape is None:
                # We will set shape, wcs from wcs_kernel on loading the first obs                
                self.rhs = None
                self.div = None
                self.hits = None
            else:
                self.init_maps_rectpix(shape, wcs)

    @classmethod
    def for_rectpix(cls, shape, wcs, comm, comps="TQU", name="sky", ofmt="{name}", output=True,
            ext="fits", dtype_map=np.float64, dtype_tod=np.float32, sys=None, recenter=None, tile_shape=(500,500), tiled=False, Nsplits=1, singlestream=False):
        """
        Signal describing a non-distributed sky map. Signal describing a sky map in the coordinate 
        system given by "sys", which defaults to equatorial coordinates. If tiled==True, then this 
        will be a distributed map with the given tile_shape, otherwise it will be a plain enmap.
        
        Arguments
        ---------
        shape : numpy.ndarray
            Shape of the output map geometry. If None, computed from coords.get_footprint on first add_obs.
        wcs : wcs
            WCS of the output map geometry (or wcs kernel).
        comm : MPI.comm
            MPI communicator
        comps : str, optional
            Components to map
        name : str, optional
            The name of this signal, e.g. "sky", "cut", etc.
        ofmt : str, optional
            The format used when constructing output file prefix
        output : Bool, optional
            Whether this signal should be part of the output or not.
        ext : str, optional
            The extension used for the files.
        dtype_map : numpy.dtype, optional
            The data type to use for the maps.
        dtype_tod : numpy.dtype, optional
            The data type to use for the time ordered data
        sys : str or None, optional
            The coordinate system to map. Defaults to equatorial
        recenter : str or None
            String to make object-centered maps, such as Moon/Sun/Planet centered maps.
            Look at sotodlib.mapmaking.parse_recentering for details.
        tile_shape : list, optional
            List with the shape of the tiles when using tiled maps from pixell
        tiled : Bool, optional
            If True, use tiled maps from pixell. If False, enmaps will be used.
        Nsplits : int, optional
            Number of splits that you will map simultaneously. By default is 1 when no
            splits are requested.
        singlestream : Bool, optional
            If True, do not perform demodulated filter+bin mapmaking but rather regular 
            filter+bin mapmaking, i.e. map from obs.signal rather than from obs.dsT, 
            obs.demodQ, obs.demodU
        
        """
        return cls(shape=shape, wcs=wcs, comm=comm, comps=comps, name=name, ofmt=ofmt, output=output,
                   ext=ext, dtype_map=dtype_map, dtype_tod=dtype_tod, sys=sys, recenter=recenter, tile_shape=tile_shape, tiled=tiled,
                   Nsplits=Nsplits, singlestream=singlestream, nside=None, nside_tile=None)

    @classmethod
    def for_healpix(cls, nside, nside_tile=None, comps="TQU", name="sky", ofmt="{name}", output=True,
            ext="fits.gz", dtype_map=np.float64, dtype_tod=np.float32, Nsplits=1, singlestream=False):
        """
        Signal describing a sky map in healpix pixelization, NEST ordering.

        Arguments
        ---------
        nside : int
            Nside of the output map. Should be a power of 2
        nside_tile: int, str or None, optional
            Nside of the tiling scheme. Should be a power of 2 and smaller than nside.
            May also be 'auto' to set automatically, or None to use no tiling.
        comps : str, optional
            Components to map
        name : str, optional
            The name of this signal, e.g. "sky", "cut", etc.
        ofmt : str, optional
            The format used when constructing output file prefix
        output : Bool, optional
            Whether this signal should be part of the output or not.
        ext : str, optional
            The extension used for the files.
            May be 'fits', 'fits.gz', 'h5', 'h5py', or 'npy'.
        dtype_map : numpy.dtype, optional
            The data type to use for the maps.
        dtype_tod : numpy.dtype, optional
            The data type to use for the time ordered data
        Nsplits : int, optional
            Number of splits that you will map simultaneously. By default is 1 when no
            splits are requested.
        singlestream : Bool, optional
            If True, do not perform demodulated filter+bin mapmaking but rather regular
            filter+bin mapmaking, i.e. map from obs.signal rather than from obs.dsT,
            obs.demodQ, obs.demodU
        """
        return cls(shape=None, wcs=None, comm=None, comps=comps, name=name, ofmt=ofmt, output=output,
                   ext=ext, dtype_map=dtype_map, dtype_tod=dtype_tod, sys=None, recenter=None, tile_shape=None, tiled=False,
                   Nsplits=Nsplits, singlestream=singlestream, nside=nside, nside_tile=nside_tile)

    def add_obs(self, id, obs, nmat, Nd, pmap=None, split_labels=None, apply_wobble=True):
        # Nd will have 3 components, corresponding to ds_T, demodQ, demodU with the noise model applied
        """Add and process an observation, building the pointing matrix
        and our part of the RHS. "obs" should be an Observation axis manager,
        nmat a noise model, representing the inverse noise covariance matrix,
        and Nd the result of applying the noise model to the detector time-ordered data.
        """
        ctime  = obs.timestamps
        if apply_wobble:
            from sotodlib.coords.helpers import get_deflected_sightline
            # Get deflection parameters
            wobble_meta = obs.wobble_params
            # Get wobble-corrected sightline
            sight = get_deflected_sightline(
                obs,
                wobble_meta,
                site=smutils.unarr(obs.site),
                weather=smutils.unarr(obs.weather)
            )
        else:
            sight = None
        for n_split in range(self.Nsplits):
            if pmap is None:
                # Build the local geometry and pointing matrix for this observation
                if self.recenter:
                    rot = smutils.recentering_to_quat_lonlat(*smutils.evaluate_recentering(self.recenter, ctime=ctime[len(ctime)//2], geom=(self.rhs.shape, self.rhs.wcs), site=smutils.unarr(obs.site)))
                else: rot = None
                if self.Nsplits == 1:
                    # this is the case with no splits
                    cuts = obs.flags.glitch_flags
                else:
                    # remember that the dets or samples you want to keep should be false, hence we negate
                    cuts = obs.flags.glitch_flags + ~obs.preprocess.split_flags.cuts[split_labels[n_split]]
                if self.pix_scheme == "rectpix":
                    threads='domdir'
                    if self.rhs is None: # Still need to initialize the geometry
                        geom = None
                        wcs_kernel = self.wcs
                    else:
                        geom = self.rhs.geometry
                        wcs_kernel = None
                else:
                    threads = ["tiles", "simple"][self.hp_geom.nside_tile is None]
                    geom = self.hp_geom
                    wcs_kernel = None
                pmap_local = coords.pmat.P.for_tod(obs, comps=self.comps, geom=geom, rot=rot, wcs_kernel=wcs_kernel, threads=threads, weather=smutils.unarr(obs.weather), site=smutils.unarr(obs.site), cuts=cuts, hwp=True, sight=sight)
            else:
                pmap_local = pmap

            if self.rhs is None: # Set the geometry now from the pmat
                shape, wcs = pmap_local.geom
                self.init_maps_rectpix(shape, wcs)
                self.wcs = wcs

            if not(self.singlestream):
                obs_rhs, obs_div, obs_hits = project_all_demod(pmap=pmap_local, signalT=obs.dsT.astype(self.dtype_tod), signalQ=obs.demodQ.astype(self.dtype_tod), signalU=obs.demodU.astype(self.dtype_tod),
                                                               det_weightsT=2*nmat.ivar, det_weightsQU=nmat.ivar, ncomp=self.ncomp, wrapper=self.wrapper)
            else:
                obs_rhs, obs_div, obs_hits = project_all_single(pmap=pmap_local, Nd=Nd, det_weights=nmat.ivar, comps='TQU', wrapper=self.wrapper)
            # Update our full rhs and div. This works for both plain and distributed maps
            if self.pix_scheme == "rectpix":
                self.rhs[n_split] = self.rhs[n_split].insert(obs_rhs, op=np.ndarray.__iadd__)
                self.div[n_split] = self.div[n_split].insert(obs_div, op=np.ndarray.__iadd__)
                self.hits[n_split] = self.hits[n_split].insert(obs_hits[0],op=np.ndarray.__iadd__)
                obs_geo = obs_rhs.geometry
            else:
                self.rhs[n_split] = obs_rhs
                self.div[n_split] = obs_div
                self.hits[n_split] = obs_hits[0]
                obs_geo = self.hp_geom

            # Save the per-obs things we need. Just the pointing matrix in our case.
            # Nmat and other non-Signal-specific things are handled in the mapmaker itself.
            self.data[(id,n_split)] = bunch.Bunch(pmap=pmap_local, obs_geo=obs_geo)
        del Nd

    def prepare(self):
        """Called when we're done adding everything. Sets up the map distribution,
        degrees of freedom and preconditioner."""
        if self.pix_scheme == "healpix": return
        if self.ready: return
        if self.tiled:
            self.geo_work = self.rhs.geometry
            self.rhs  = tilemap.redistribute(self.rhs, self.comm)
            self.div  = tilemap.redistribute(self.div, self.comm)
            self.hits = tilemap.redistribute(self.hits,self.comm)
        else:
            if self.comm is not None:
                self.rhs  = putils.allreduce(self.rhs, self.comm)
                self.div  = putils.allreduce(self.div, self.comm)
                self.hits = putils.allreduce(self.hits,self.comm)
        self.ready = True

    @property
    def ncomp(self): return len(self.comps)

    def write(self, prefix, tag, m, unit='K'):
        if not self.output: return
        oname = self.ofmt.format(name=self.name)
        oname = "%s%s_%s.%s" % (prefix, oname, tag, self.ext)
        if self.pix_scheme == "rectpix":
            if self.tiled:
                tilemap.write_map(oname, m, self.comm, extra={'BUNIT':unit})
            else:
                if self.comm is None or self.comm.rank == 0:
                    enmap.write_map(oname, m, extra={'BUNIT':unit})
        else:
            if self.ext in ["fits", "fits.gz"]:
                if hp is None:
                    raise ImportError("Cannot save healpix map as fits; healpy could not be imported. Install healpy or save as npy or h5")
                if m.ndim > 2:
                    m = np.reshape(m, (np.prod(m.shape[:-1]), m.shape[-1]), order='C') # Flatten wrapping axes; healpy.write_map can't handle >2d array
                hp.write_map(oname, m.view(self.dtype_map), nest=(self.hp_geom.ordering=='NEST'), overwrite=True, column_units=unit)
            elif self.ext == "npy":
                np.save(oname, {'nside': self.hp_geom.nside, 'ordering':self.hp_geom.ordering, 'data':m.view(self.dtype_map)}, allow_pickle=True)
            elif self.ext in ['h5', 'hdf5']:
                if h5py is None:
                    raise ValueError("Cannot save healpix map as hdf5; h5py could not be imported. Install h5py or save as npy or fits")
                with h5py.File(oname, 'w') as f:
                    dset = f.create_dataset("data", m.shape, dtype=self.dtype_map, data=m)
                    dset.attrs['ordering'] = self.hp_geom.ordering
                    dset.attrs['nside'] = self.hp_geom.nside
            else:
                raise ValueError(f"Unknown extension {self.ext}")

        return oname

    def init_maps_rectpix(self, shape, wcs):
        """ Initialize tilemaps or enmaps rhs, div, hits for given shape and wcs"""
        shape = tuple(shape[-2:])
        Nsplits, ncomp, dtype = self.Nsplits, self.ncomp, self.dtype_map
        
        if self.tiled:
            geo = tilemap.geometry(shape, wcs, tile_shape=self.tile_shape)
            rhs = tilemap.zeros(geo.copy(pre=(Nsplits,ncomp,)),      dtype=self.dtype_map)
            div = tilemap.zeros(geo.copy(pre=(Nsplits,ncomp,ncomp)), dtype=self.dtype_map)
            hits= tilemap.zeros(geo.copy(pre=(Nsplits,)),            dtype=self.dtype_map)
        else:
            rhs = enmap.zeros((Nsplits, ncomp)     +shape, wcs, dtype=self.dtype_map)
            div = enmap.zeros((Nsplits,ncomp,ncomp)+shape, wcs, dtype=self.dtype_map)
            hits= enmap.zeros((Nsplits,)+shape, wcs, dtype=self.dtype_map)
        self.rhs = rhs
        self.div = div
        self.hits = hits
        return rhs, div, hits
    

def setup_demod_map(noise_model, shape=None, wcs=None, nside=None,
                    comm=mpi.COMM_WORLD, comps='TQU', split_labels=['full'],
                    singlestream=False, dtype_tod=np.float32,
                    dtype_map=np.float64, recenter=None, verbose=0):
    """
    Setup the classes for demod mapmaking and return
    a DemodMapmmaker object
    """
    if wcs is not None:
        Nsplits = len(split_labels)
        signal_map = DemodSignalMap.for_rectpix(shape, wcs, comm, comps=comps,
                                              dtype_map=dtype_map, dtype_tod=dtype_tod, tiled=False,
                                              ofmt="", Nsplits=Nsplits,
                                              singlestream=singlestream,
                                              recenter=recenter)
    elif nside is not None:
        Nsplits = len(split_labels)
        signal_map = DemodSignalMap.for_healpix(nside, nside_tile='auto', 
                                    comps=comps, dtype_map=dtype_map, dtype_tod=dtype_tod,
                                    ofmt="", Nsplits=Nsplits,
                                    singlestream=singlestream,
                                    ext="fits.gz")
    signals    = [signal_map]
    mapmaker   = DemodMapmaker(signals, noise_model=noise_model,
                                         dtype=dtype_tod,
                                         verbose=verbose>0,
                                         singlestream=singlestream)
    return mapmaker

def write_demod_maps(prefix, data, info, unit='K', split_labels=['full']):
    """
    Write maps from data into files

    Arguments
    ---------
    prefix : str
        Prefix where to write maps.
    data : Bunch
        Bunch object with required maps.
    info : list
        List of dictionaries with the info to write into the atomic
        db. Will be modified in place for valid=[True|False] atomics.
    unit : str, optional
        Unit to write into the header of the fits file.
    split_labels : list, optional
        List with splits labels.
    """
    Nsplits = len(split_labels)
    for n_split in range(Nsplits):
        if np.all(data.wmap[n_split] == 0.0):
            info[n_split]['valid'] = False
        else:
            info[n_split]['valid'] = True
            data.signal.write(prefix, "%s_wmap"%split_labels[n_split],
                              data.wmap[n_split], unit=unit+'^-1')
            data.signal.write(prefix, "%s_weights"%split_labels[n_split],
                              data.weights[n_split], unit=unit+'^2')
            data.signal.write(prefix, "%s_hits"%split_labels[n_split],
                              data.signal.hits[n_split], unit='hits')

def make_demod_map(context, obslist, noise_model, info,
                    preprocess_config, prefix, shape=None, wcs=None,
                    nside=None, comm=mpi.COMM_WORLD, comps="TQU", t0=0,
                    dtype_tod=np.float32, dtype_map=np.float32,
                    tag="", verbose=0, split_labels=['full'], L=None,
                    site='so_sat3', recenter=None, singlestream=False,
                    unit='K', use_psd=True, wn_label='preprocess.noiseQ_mapmaking.white_noise',
                    apply_wobble=True):
    """
    Make a demodulated map from the list of observations in obslist.

    Arguments
    ---------
    context : str
        File path to context used to load obs from.
    obslist : dict
        The obslist which is the output of the
        mapmaking.obs_grouping.build_obslists, contains the information of the
        single or multiple obs to map.
    noise_model : sotodlib.mapmaking.Nmat
        Noise model to pass to DemodMapmaker.
    info : list
        Information for the database in the form of a list of dictionaries
    preprocess_config : list of dict
        List of dictionaries with the config yaml file for the preprocess database.
        If two, then a multilayer preprocessing is to be used.
    prefix : str
        Prefix for the output files
    shape : tuple, optional
        Shape of the geometry to use for mapping.
    wcs : dict, optional
        WCS kernel of the geometry to use for mapping.
    nside : int, optional
        Nside for healpix pixelization
    comps : str, optional
        Which components to map, only TQU supported for now.
    t0 : int, optional
        Ctime to use as the label in the map files.
    dtype_tod : numpy.dtype, optional
        The data type to use for the time-ordered data. Only tested
        with float32.
    dtype_map : numpy.dtype, optional
        The data type to use for the maps.
    tag : str, optional
        Prefix tag for the logger.
    verbose : bool, optional
    split_labels : list or None, optional
        A list of strings with the splits requested. If None then no splits
        were asked for, i.e. we will produce one map.
    L : logger, optional
        Logger for printing on the screen.
    site : str, optional
        Platform name for the pointing matrix.
    recenter : str or None
        String to make object-centered maps, such as Moon/Sun/Planet centered maps.
        Look at sotodlib.mapmaking.parse_recentering for details.
    singlestream : Bool
        If True, do not perform demodulated filter+bin mapmaking but
        rather regular filter+bin mapmaking, i.e. map from obs.signal
        rather than from obs.dsT, obs.demodQ, obs.demodU.
    use_psd : bool, optional
        By default True. Use the white noise measured from PSD as mapmaking weights,
        which must be provided in the preprocessing. This is done as opposed to
        build the ivar locally from the std of the TOD.
    wn_label : str, optional
        Path where to find the white noise per det estimated by the preprocessing.
    apply_wobble : bool, optional
        Correct wobble deflection. This requires aman.wobble_params metadata in the
        context

    Returns
    -------
    errors : list
        List of errors from preprocess database. To be used in cleanup_mandb.
    outputs : list
        List of outputs from preprocess database. To be used in cleanup_mandb.
    info: list
        List of dictionaries with the info to write into atomic db. It will be
        the same as the input, but the weights values will be added.
    """
    from ..preprocess import preprocess_util
    #context = core.Context(context)
    if L is None:
        L = preprocess_util.init_logger("Demod filterbin mapmaking")
    pre = "" if tag is None else tag + " "
    if comm.rank == 0: L.info(pre + "Initializing equation system")
    mapmaker = setup_demod_map(noise_model, shape=shape, wcs=wcs, nside=nside,
                    comm=comm, comps=comps, split_labels=split_labels,
                    singlestream=singlestream, dtype_tod=dtype_tod,
                    dtype_map=dtype_map, recenter=recenter, verbose=verbose)

    if comm.rank == 0: L.info(pre + "Building RHS")
    # And feed it with our observations
    nobs_kept  = 0
    errors = [] ; outputs = []; # PENDING: do an allreduce of these.
                                # not needed for atomic maps, but needed for
                                # depth-1 maps
    if len(preprocess_config)==1:
        preproc_init = preprocess_config[0]
        preproc_proc = None
    else:
        preproc_init = preprocess_config[0]
        preproc_proc = preprocess_config[1]
    n_dets = 0
    for oi in range(len(obslist)):
        obs_id, detset, band = obslist[oi][:3]
        name = "%s:%s:%s" % (obs_id, detset, band)
        error, output_init, output_proc, obs = preprocess_util.preproc_or_load_group(obs_id,
                                                configs_init=preproc_init,
                                                configs_proc=preproc_proc,
                                                dets={'wafer_slot':detset, 'wafer.bandpass':band},
                                                logger=L,
                                                overwrite=False)
        errors.append(error) ; outputs.append((output_init, output_proc)) ;
        if error not in [None,'load_success']:
            L.info('tod %s:%s:%s failed in the preproc database'%(obs_id,detset,band))
            continue
        obs.wrap("weather", np.full(1, "toco"))
        obs.wrap("site",    np.full(1, site))
        mapmaker.add_obs(name, obs, split_labels=split_labels, use_psd=use_psd, wn_label=wn_label, apply_wobble=apply_wobble)
        L.info('Done with tod %s:%s:%s'%(obs_id,detset,band))
        nobs_kept += 1
        n_dets += obs.dets.count
    nobs_kept = comm.allreduce(nobs_kept)
    n_dets = comm.allreduce(n_dets)
    for subinfo in info:
        subinfo['number_dets'] = n_dets
    # if we skip all the obs then we return error and output
    if nobs_kept == 0:
        return errors, outputs, None

    for signal in mapmaker.signals:
        signal.prepare()
    if comm.rank == 0: L.info(pre + "Writing F+B outputs")
    wmap = []
    weights = []
    # mapmaker.signals[0] is signal_map
    for n_split in range(mapmaker.signals[0].Nsplits):
        wmap.append( mapmaker.signals[0].rhs[n_split] )
        div = np.diagonal(mapmaker.signals[0].div[n_split], axis1=0, axis2=1)
        div = np.moveaxis(div, -1, 0) # this moves the last axis to the 0th position
        weights.append(div)
    mapdata = bunch.Bunch(wmap=wmap, weights=weights, signal=mapmaker.signals[0], t0=t0)
    info = add_weights_to_info(info, weights, split_labels)

    # output to files
    write_demod_maps(prefix, mapdata, info, split_labels=split_labels, unit=unit)
    return errors, outputs , info

def add_weights_to_info(info, weights, split_labels):
    Nsplits = len(split_labels)
    for isplit in range(Nsplits):
        sub_info = info[isplit]
        sub_weights = weights[isplit]
        # Assuming weights are TT, QQ, UU
        if sub_weights.shape[0] != 3:
            raise ValueError(f"sub_weights has unexpected shape {sub_weights.shape}. First axis should be (3,) for TT, QQ, UU")
        mean_qu = np.mean(sub_weights[1:], axis=0)
        pweights = mean_qu[mean_qu > 0]
        if pweights.size == 0:
            pweights = [0]
        sumweights = np.sum(pweights)
        meanweights = np.mean(pweights)
        medianweights = np.median(pweights)
        sub_info['total_weight_qu'] = sumweights
        sub_info['mean_weight_qu'] = meanweights
        sub_info['median_weight_qu'] = medianweights
        info[isplit] = sub_info
    return info

def project_rhs_demod(pmap, signalT, signalQ, signalU, det_weightsT, det_weightsQU, wrapper=lambda x:x):
    """
    Project demodulated T, Q, U timestreams into weighted maps.

    Arguments
    ---------
    pmap : sotodlib.coords.pmat.P
        Projection matrix.
    signalT, signalQ, signalU : np.ndarray (ndets, nsamps)
        T, Q, U timestreams after demodulation.
    det_weightsT, det_weightsQU : np.ndarray (ndets,) or None
        Array of detector weights for T and QU. If None, unit weights are used.
    wrapper : Function with single argument
        Wrapper function for output of pmap operations.
    """
    zeros = lambda *args, **kwargs : wrapper(pmap.zeros(*args, **kwargs))
    to_map = lambda *args, **kwargs : wrapper(pmap.to_map(*args, **kwargs))

    rhs = zeros()
    rhs_T = to_map(signal=signalT, comps='T', det_weights=det_weightsT)
    rhs_demodQ = to_map(signal=signalQ, comps='QU', det_weights=det_weightsQU)
    rhs_demodU = to_map(signal=signalU, comps='QU', det_weights=det_weightsQU)
    rhs_demodQU = zeros(super_shape=(2), comps='QU')

    rhs_demodQU[0][:] = rhs_demodQ[0] - rhs_demodU[1]
    rhs_demodQU[1][:] = rhs_demodQ[1] + rhs_demodU[0]
    del rhs_demodQ, rhs_demodU

    # we write into the rhs.
    rhs[0] = rhs_T[0]
    rhs[1] = rhs_demodQU[0]
    rhs[2] = rhs_demodQU[1]
    return rhs

def project_div_demod(pmap, det_weightsT, det_weightsQU, ncomp, wrapper=lambda x:x):
    """
    Make weight maps for demodulated data.

    Arguments
    ---------
    pmap : sotodlib.coords.pmat.P
        Projection matrix.
    det_weightsT, det_weightsQU : np.ndarray (ndets,) or None
        Array of detector weights for T and QU. If None, unit weights are used.
    ncomp : int
        Number of map components. e.g. 3 for TQU.
    wrapper : Function with single argument
        Wrapper function for output of pmap operations.
    """
    zeros = lambda *args, **kwargs : wrapper(pmap.zeros(*args, **kwargs))
    to_weights = lambda *args, **kwargs : wrapper(pmap.to_weights(*args, **kwargs))

    div = zeros(super_shape=(ncomp, ncomp))
    # Build the per-pixel inverse covmat for this observation
    wT = to_weights(comps='T', det_weights=det_weightsT)
    wQU = to_weights(comps='T', det_weights=det_weightsQU)
    div[0,0] = wT
    div[1,1] = wQU
    div[2,2] = wQU
    return div

def project_all_demod(pmap, signalT, signalQ, signalU, det_weightsT, det_weightsQU, ncomp, wrapper=lambda x:x):
    """
    Get weighted signal, weight, and hits maps for demodulated data.
    See project_rhs_demod for description of args.
    """
    rhs =  project_rhs_demod(pmap, signalT, signalQ, signalU, det_weightsT, det_weightsQU, wrapper)
    div = project_div_demod(pmap, det_weightsT, det_weightsQU, ncomp, wrapper)
    hits = wrapper(pmap.to_map(signal=np.ones_like(signalT))) ## Note hits is *not* weighted by det_weights
    return rhs, div, hits

def project_all_single(pmap, Nd, det_weights, comps, wrapper=lambda x:x):
    """
    Get weighted signal, weight and hits maps from a single (non-demodulated) timestream.

    Arguments
    ---------
    pmap : sotodlib.coords.pmat.P
        Projection matrix.
    Nd : np.ndarray (ndets, nsamps)
        Input timestream to map.
    det_weights : np.ndarray (ndets,) or None
        Array of detector weights. If None, unit weights are used.
    comps : str
        Components to use in mapmaking. 'TQU', 'T', or 'QU'.
    wrapper : Function with single argument
        Wrapper function for output of pmap operations.
    """
    ncomp = len(comps)
    rhs = wrapper(pmap.to_map(signal=Nd, comps=comps, det_weights=det_weights))
    div = pmap.zeros(super_shape=(ncomp, ncomp))
    pmap.to_weights(dest=div, comps=comps, det_weights=det_weights)
    div = wrapper(div)
    hits = wrapper(pmap.to_map(signal=np.ones_like(Nd)))
    return rhs, div, hits



# Copyright (c) 2018-2019 Simons Observatory.
# Full license can be found in the top level "LICENSE" file.
"""Test Suite for Simons Observatory TOD Simulation and Processing.

This directory contains the unittest suite.

"""



# Copyright (c) 2018-2019 Simons Observatory.
# Full license can be found in the top level "LICENSE" file.
"""Trim a hardware model to include only some detectors.
"""

import sys
import argparse

from collections import OrderedDict

from ..core import Hardware


def main():
    parser = argparse.ArgumentParser(
        description="This program reads a hardware model from disk,\
            selects a subset of detectors, and writes the new model out.",
        usage="so_hardware_trim [options] (use --help for details)")

    parser.add_argument(
        "--hardware", required=True, default=None,
        help="Input hardware file"
    )

    parser.add_argument(
        "--out", required=False, default="trimmed",
        help="Name (without extensions) of the output hardware file"
    )

    parser.add_argument(
        "--plain", required=False, default=False, action="store_true",
        help="Write plain text (without gzip compression)"
    )

    parser.add_argument(
        "--overwrite", required=False, default=False, action="store_true",
        help="Overwrite any existing output file."
    )

    parser.add_argument(
        "--telescopes", required=False, default=None,
        help="Select only detectors on these telescope (LAT, SAT0, SAT1, etc)\
            .  This should be either a regex string or a comma-separated\
            list of names."
    )

    parser.add_argument(
        "--tube_slots", required=False, default=None,
        help="Select only detectors on these tube slots.  This should be either a\
            regex string or a comma-separated list of names."
    )

    parser.add_argument(
        "--match", required=False, default=None, nargs="*",
        help="Specify one or more detector match criteria.  Each match should\
            have the format '<property>:<regex or list>'.  The regex\
            expression should be valid to pass to the 're' module.  If \
            passing a list, this should be comma-separated.  For example, \
            --match 'band:SAT_MF.*' 'wafer_slot:w25,w26' 'pol:A' "
    )

    args = parser.parse_args()

    telescopes = args.telescopes
    if telescopes is not None:
        telescopes = telescopes.split(",")
    tube_slots = args.tube_slots
    if tube_slots is not None:
        tube_slots = tube_slots.split(",")

    match = dict()
    if args.match is not None:
        for amat in args.match:
            fields = amat.split(":")
            if len(fields) != 2:
                raise ValueError(
                    "Invalid match string- must have a colon separator.")
            temp = fields[1].split(",")
            if len(temp) > 1:
                match[fields[0]] = temp
            else:
                match[fields[0]] = fields[1]

    print("Loading hardware from {}...".format(args.hardware), flush=True)
    hw = Hardware(args.hardware)

    print("Selecting detectors from:")
    if telescopes is not None:
        telstr = ""
        for tele in telescopes:
            telstr = "{}'{}', ".format(telstr, tele)
        print("  telescopes = {}".format(telstr.rstrip(", ")), flush=True)
    if tube_slots is not None:
        tstr = ""
        for tube_slot in tube_slots:
            tstr = "{}'{}', ".format(tstr, tube_slot)
        print("  tube_slots = {}".format(tstr.rstrip(", ")), flush=True)
    for k, v in match.items():
        print("  {} = r'{}'".format(k, v), flush=True)

    newhw = hw.select(telescopes=telescopes, tube_slots=tube_slots, match=match)

    if args.plain:
        outpath = "{}.toml".format(args.out)
        print("Dumping selected config to {}...".format(outpath))
        newhw.dump(outpath, overwrite=args.overwrite, compress=False)
    else:
        outpath = "{}.toml.gz".format(args.out)
        print("Dumping selected config to {}...".format(outpath))
        newhw.dump(outpath, overwrite=args.overwrite, compress=True)

    return



# Copyright (c) 2023-2023 Simons Observatory.
# Full license can be found in the top level "LICENSE" file.
"""Simulated detector response to calibrators.
"""

import numpy as np
from astropy import units as u
import toast
import toast.ops

from .. import ops as so_ops
from .job import workflow_timer


def setup_simulate_wiregrid_signal(operators):
    """Add commandline args and operators for simulating wiregrid.

    Args:
        operators (list):  The list of operators to extend.

    Returns:
        None

    """
    operators.append(so_ops.SimWireGrid(name="sim_wiregrid", enabled=False))


@workflow_timer
def simulate_wiregrid_signal(job, otherargs, runargs, data):
    """Simulate wire grid signal.

    Args:
        job (namespace):  The configured operators and templates for this job.
        otherargs (namespace):  Other commandline arguments.
        runargs (namespace):  Job related runtime parameters.
        data (Data):  The data container.

    Returns:
        None

    """
    # Configured operators for this job
    job_ops = job.operators

    if job_ops.sim_wiregrid.enabled:
        job_ops.sim_wiregrid.detector_pointing = job_ops.det_pointing_azel
        job_ops.sim_wiregrid.detector_weights = job_ops.weights_azel
        job_ops.sim_wiregrid.apply(data)


def setup_simulate_stimulator_signal(operators):
    """Add commandline args and operators for simulating the stimulator.

    Args:
        operators (list):  The list of operators to extend.

    Returns:
        None

    """
    operators.append(so_ops.SimStimulator(name="sim_stimulator", enabled=False))


@workflow_timer
def simulate_stimulator_signal(job, otherargs, runargs, data):
    """Simulate stimulator signal.

    Args:
        job (namespace):  The configured operators and templates for this job.
        otherargs (namespace):  Other commandline arguments.
        runargs (namespace):  Job related runtime parameters.
        data (Data):  The data container.

    Returns:
        None

    """
    # Configured operators for this job
    job_ops = job.operators

    if job_ops.sim_stimulator.enabled:
        job_ops.sim_stimulator.apply(data)
